@@ -10,13 +10,15 @@ import sys
 from shutil import copytree
 from optparse import OptionParser
 
-from gi.repository import GLib, Gtk
+from gi.repository import GLib, Gtk, Gio
 
-from Onboard.utils        import show_confirmation_dialog, Version, unicode_str
-from Onboard.WindowUtils  import Handle
-from Onboard.ConfigUtils  import ConfigObject
-from Onboard.MouseControl import Mousetweaks, ClickMapper
-from Onboard.Exceptions   import SchemaError
+from Onboard.utils          import show_confirmation_dialog, Version, \
+                                   unicode_str, XDGDirs, chmodtree
+from Onboard.definitions    import InputEventSourceEnum, TouchInputEnum
+from Onboard.WindowUtils    import Handle, DockingEdge
+from Onboard.ConfigUtils    import ConfigObject
+from Onboard.ClickSimulator import CSMousetweaks
+from Onboard.Exceptions     import SchemaError
 
 ### Logging ###
 import logging
@@ -24,23 +26,25 @@ _logger = logging.getLogger("Config")
 ###############
 
 # gsettings schemas
-SCHEMA_ONBOARD          = "org.onboard"
-SCHEMA_KEYBOARD         = "org.onboard.keyboard"
-SCHEMA_WINDOW           = "org.onboard.window"
-SCHEMA_WINDOW_LANDSCAPE = "org.onboard.window.landscape"
-SCHEMA_WINDOW_PORTRAIT  = "org.onboard.window.portrait"
-SCHEMA_ICP              = "org.onboard.icon-palette"
-SCHEMA_ICP_LANDSCAPE    = "org.onboard.icon-palette.landscape"
-SCHEMA_ICP_PORTRAIT     = "org.onboard.icon-palette.portrait"
-SCHEMA_AUTO_SHOW        = "org.onboard.auto-show"
-SCHEMA_UNIVERSAL_ACCESS = "org.onboard.universal-access"
-SCHEMA_THEME            = "org.onboard.theme-settings"
-SCHEMA_LOCKDOWN         = "org.onboard.lockdown"
-SCHEMA_SCANNER          = "org.onboard.scanner"
+SCHEMA_ONBOARD           = "org.onboard"
+SCHEMA_KEYBOARD          = "org.onboard.keyboard"
+SCHEMA_WINDOW            = "org.onboard.window"
+SCHEMA_WINDOW_LANDSCAPE  = "org.onboard.window.landscape"
+SCHEMA_WINDOW_PORTRAIT   = "org.onboard.window.portrait"
+SCHEMA_ICP               = "org.onboard.icon-palette"
+SCHEMA_ICP_LANDSCAPE     = "org.onboard.icon-palette.landscape"
+SCHEMA_ICP_PORTRAIT      = "org.onboard.icon-palette.portrait"
+SCHEMA_AUTO_SHOW         = "org.onboard.auto-show"
+SCHEMA_UNIVERSAL_ACCESS  = "org.onboard.universal-access"
+SCHEMA_THEME             = "org.onboard.theme-settings"
+SCHEMA_LOCKDOWN          = "org.onboard.lockdown"
+SCHEMA_SCANNER           = "org.onboard.scanner"
+SCHEMA_TYPING_ASSISTANCE = "org.onboard.typing-assistance"
+SCHEMA_WORD_SUGGESTIONS  = "org.onboard.typing-assistance.word-suggestions"
 
-SCHEMA_GSS              = "org.gnome.desktop.screensaver"
-SCHEMA_GDI              = "org.gnome.desktop.interface"
-SCHEMA_GDA              = "org.gnome.desktop.a11y.applications"
+SCHEMA_GSS               = "org.gnome.desktop.screensaver"
+SCHEMA_GDI               = "org.gnome.desktop.interface"
+SCHEMA_GDA               = "org.gnome.desktop.a11y.applications"
 
 MODELESS_GKSU_KEY = "/apps/gksu/disable-grab"  # old gconf key, unused
 
@@ -50,6 +54,10 @@ DEFAULT_Y                  = 50    # else dconf data migration won't happen.
 DEFAULT_WIDTH              = 700
 DEFAULT_HEIGHT             = 205
 
+# Default rect on Nexus 7
+# landscape x=65, y=500, w=1215 h=300
+# portrait  x=55, y=343, w=736 h=295
+
 DEFAULT_ICP_X              = 100   # Make sure these match the schema defaults,
 DEFAULT_ICP_Y              = 50    # else dconf data migration won't happen.
 DEFAULT_ICP_HEIGHT         = 64
@@ -61,20 +69,20 @@ DEFAULT_COLOR_SCHEME       = "Classic Onboard"
 
 START_ONBOARD_XEMBED_COMMAND = "onboard --xid"
 
-GTK_KBD_MIXIN_MOD          = "Onboard.KeyboardGTK"
-GTK_KBD_MIXIN_CLS          = "KeyboardGTK"
-
 INSTALL_DIR                = "/usr/share/onboard"
 LOCAL_INSTALL_DIR          = "/usr/local/share/onboard"
-USER_DIR                   = ".onboard"
+USER_DIR                   = "onboard"
 
 SYSTEM_DEFAULTS_FILENAME   = "onboard-defaults.conf"
 
 DEFAULT_RESIZE_HANDLES     = list(Handle.RESIZERS)
 
+DEFAULT_FREQUENCY_TIME_RATIO = 75  # 0=100% frequency, 100=100% time (last use)
+
 SCHEMA_VERSION_0_97         = Version(1, 0)   # Onboard 0.97
-SCHEMA_VERSION_0_98       = Version(2, 0)   # Onboard 0.97.1
-SCHEMA_VERSION              = SCHEMA_VERSION_0_98
+SCHEMA_VERSION_0_98         = Version(2, 0)   # Onboard 0.97.1
+SCHEMA_VERSION_0_99         = Version(2, 1)   # Onboard 0.99.0
+SCHEMA_VERSION              = SCHEMA_VERSION_0_99
 
 
 # enum for simplified number of resize_handles
@@ -83,19 +91,11 @@ class NumResizeHandles:
     SOME = 1
     ALL  = 2
 
-
 class Config(ConfigObject):
     """
     Singleton Class to encapsulate the gsettings stuff and check values.
     """
 
-    # String representation of the module containing the Keyboard mixin
-    # used to draw keyboard
-    _kbd_render_mixin_mod = GTK_KBD_MIXIN_MOD
-
-    # String representation of the keyboard mixin used to draw keyboard.
-    _kbd_render_mixin_cls = GTK_KBD_MIXIN_CLS
-
     # extension of layout files
     LAYOUT_FILE_EXTENSION = ".onboard"
 
@@ -127,6 +127,16 @@ class Config(ConfigObject):
     # raised border size of dish keys
     DISH_KEY_BORDER = (2.5, 2.5)
 
+    # minimum time keys are drawn in pressed state
+    UNPRESS_DELAY = 0.15
+
+    # Margin to leave around wordlist labels; smaller margins leave
+    # room for more prediction choices
+    WORDLIST_LABEL_MARGIN = (2, 2)
+
+    # Gap between wordlist buttons
+    WORDLIST_BUTTON_SPACING = (1, 1)
+
     # index of currently active pane, not stored in gsettings
     active_layer_index = 0
 
@@ -145,7 +155,7 @@ class Config(ConfigObject):
         Singleton magic.
         """
         if not hasattr(cls, "self"):
-            cls.self = object.__new__(cls, args, kwargs)
+            cls.self = object.__new__(cls, *args, **kwargs)
             cls.self.construct()
         return cls.self
 
@@ -187,11 +197,22 @@ class Config(ConfigObject):
         parser.add_option("-q", "--quirks", dest="quirks",
                 help=_("Override auto-detection and manually select quirks\n"
                        "QUIRKS={metacity|compiz|mutter}"))
+        parser.add_option("--not-show-in", dest="not_show_in",
+                metavar="DESKTOPS",
+                help=_("Silently fail to start in the given desktop "
+                       "environments. DESKTOPS is a comma-separated list of "
+                       "XDG desktop names, e.g. GNOME for GNOME Shell."
+                       ))
+        parser.add_option("-g", "--log-learning",
+                  action="store_true", dest="log_learn", default=False,
+                  help="log all learned text; off by default")
+
         options = parser.parse_args()[0]
         self.options = options
 
         self.xid_mode = options.xid_mode
         self.quirks = options.quirks
+        self.log_learn = options.log_learn
 
         # setup logging
         log_params = {
@@ -201,8 +222,9 @@ class Config(ConfigObject):
              log_params["level"] = getattr(logging, options.debug.upper())
         if False: # log to file
             log_params["level"] = "DEBUG"
-            log_params["filename"] = "/tmp/onboard.log"
-            log_params["filemode"] = "w"
+            logfile = open("/tmp/onboard.log", "w")
+            sys.stdout = logfile
+            sys.stderr = logfile
 
         logging.basicConfig(**log_params)
 
@@ -230,47 +252,66 @@ class Config(ConfigObject):
         self.install_dir = self._get_install_dir()
         self.user_dir = self._get_user_dir()
 
-        # migrate old user dir ".sok" to ".onboard"
-        old_user_dir = os.path.join(os.path.expanduser("~"), ".sok")
-        user_dir = self.user_dir
-        if not os.path.exists(user_dir) and os.path.exists(old_user_dir):
-            _logger.info(_format("Migrating user directory '{}' to '{}'.", \
-                                 old_user_dir, user_dir))
-            try:
-                copytree(old_user_dir, user_dir)
-            except OSError as ex: # python >2.5
-                _logger.error(_format("Failed to migrate user directory. ") + \
-                              unicode_str(ex))
+        # migrate old user dir ".sok" or ".onboard" to XDG data home
+        if not os.path.exists(self.user_dir):
+            old_user_dir = os.path.join(os.path.expanduser("~"), ".onboard")
+            if not os.path.exists(old_user_dir):
+                old_user_dir = os.path.join(os.path.expanduser("~"), ".sok")
+            if os.path.exists(old_user_dir):
+                _logger.info(_format("Migrating user directory '{}' to '{}'.", \
+                                     old_user_dir, self.user_dir))
+                # Copy the data directory
+                try:
+                    copytree(old_user_dir, self.user_dir)
+                    chmodtree(self.user_dir, 0o700, True) # honor XDG spec
+
+                except OSError as ex: # python >2.5
+                    _logger.error(_format("failed to migrate user directory. ") + \
+                                  unicode_str(ex))
+
+                # Return paths in gsettings to basenames. Custom path
+                # information will be lost.
+                try:
+                    filter = lambda x: os.path.splitext(os.path.basename(x))[0]
+                    s = Gio.Settings.new(SCHEMA_ONBOARD)
+                    s["layout"] = filter(s["layout"])
+                    s["theme"] = filter(s["theme"])
+                    s["system-theme-associations"] = \
+                        {k : filter(v) \
+                         for k, v in s["system-theme-associations"].items()}
+                except Exception as ex:
+                    _logger.warning("migrating gsettings paths failed: " + \
+                                    unicode_str(ex))
 
         # Load system defaults (if there are any, not required).
         # Used for distribution specific settings, aka branding.
-        paths = [os.path.join(self.install_dir, SYSTEM_DEFAULTS_FILENAME),
+        paths = XDGDirs.get_all_config_dirs(USER_DIR) + \
+                [os.path.join(self.install_dir, SYSTEM_DEFAULTS_FILENAME),
                  os.path.join("/etc/onboard", SYSTEM_DEFAULTS_FILENAME)]
         self.load_system_defaults(paths)
 
         # initialize all property values
-        self.init_properties(self.options)
+        used_system_defaults = self.init_properties(self.options)
 
         # Make sure there is a 'Default' entry when tracking the system theme.
-        # 'Default' is the theme used when encountering an so far unknown
-        # gtk-theme. 'Default' is added on first start and therefore a
-        # corresponding system default is respected.
+        # 'Default' is the theme used when encountering a so far unknown
+        # gtk-theme.
         theme_assocs = self.system_theme_associations.copy()
         if not "Default" in theme_assocs:
-            theme_assocs["Default"] = self.theme
+            theme_assocs["Default"] = ""
             self.system_theme_associations = theme_assocs
 
-        # remember command line theme for system theme tracking
+        # Remember command line theme for system theme tracking.
         if self.options.theme:
+            self.theme_key.modified = True # force writing on next occasion
             self.remember_theme(self.theme)
 
         # load theme
         global Theme
         from Onboard.Appearance import Theme
-        self.apply_theme()
-
-        # give gtk theme a chance to take over
-        self.update_theme_from_system_theme()
+        _logger.info("Theme candidates {}" \
+                     .format(self._get_theme_candidates()))
+        self.load_theme()
 
         # misc initializations
         self._last_snippets = dict(self.snippets)  # store a copy
@@ -290,6 +331,10 @@ class Config(ConfigObject):
         # tell config objects that their properties are valid now
         self.on_properties_initialized()
 
+        # Work around changes in preferences having no effect in Saucy.
+        # If there is an unbalanced self.delay() somewhere I haven't found it.
+        self.apply()
+
         _logger.debug("Leaving init")
 
     def cleanup(self):
@@ -297,7 +342,6 @@ class Config(ConfigObject):
         # when restarting. Restarts don't happen anymore, keep
         # this for now anyway.
         self.disconnect_notifications()
-        self.clickmapper.cleanup()
         if self.mousetweaks:
             self.mousetweaks.cleanup()
 
@@ -336,16 +380,17 @@ class Config(ConfigObject):
         self.add_key("key-label-overrides", {}, "as") # default labels for all themes
         self.add_key("current-settings-page", 0)
 
-        self.keyboard         = ConfigKeyboard()
-        self.window           = ConfigWindow()
-        self.icp              = ConfigICP(self)
-        self.auto_show        = ConfigAutoShow()
-        self.universal_access = ConfigUniversalAccess(self)
-        self.theme_settings   = ConfigTheme(self)
-        self.lockdown         = ConfigLockdown(self)
-        self.scanner          = ConfigScanner(self)
-        self.gss              = ConfigGSS(self)
-        self.gdi              = ConfigGDI(self)
+        self.keyboard          = ConfigKeyboard()
+        self.window            = ConfigWindow()
+        self.icp               = ConfigICP(self)
+        self.auto_show         = ConfigAutoShow()
+        self.universal_access  = ConfigUniversalAccess(self)
+        self.theme_settings    = ConfigTheme(self)
+        self.lockdown          = ConfigLockdown(self)
+        self.scanner           = ConfigScanner(self)
+        self.typing_assistance = ConfigTypingAssistance(self)
+        self.gss               = ConfigGSS(self)
+        self.gdi               = ConfigGDI(self)
 
         self.children += [self.keyboard,
                           self.window,
@@ -356,17 +401,16 @@ class Config(ConfigObject):
                           self.lockdown,
                           self.gss,
                           self.gdi,
-                          self.scanner]
+                          self.scanner,
+                          self.typing_assistance]
 
         try:
-            self.mousetweaks = Mousetweaks()
+            self.mousetweaks = CSMousetweaks()
             self.children.append(self.mousetweaks)
         except (SchemaError, ImportError) as e:
             _logger.warning(unicode_str(e))
             self.mousetweaks = None
 
-        self.clickmapper = ClickMapper()
-
     def init_from_gsettings(self):
         """
         Overloaded to migrate old dconf data to new gsettings schemas
@@ -484,17 +528,13 @@ class Config(ConfigObject):
         self.layout_notify_add(callback)
 
     def get_layout_filename(self):
-        return self._get_user_sys_filename_gs(
-             gskey                = self.layout_key,
-             user_filename_func   = lambda x: \
-                 os.path.join(self.user_dir,    "layouts", x) + \
-                 self.LAYOUT_FILE_EXTENSION,
-             system_filename_func = lambda x: \
-                 os.path.join(self.install_dir, "layouts", x) + \
-                 self.LAYOUT_FILE_EXTENSION,
-             final_fallback       = os.path.join(self.install_dir,
-                                                "layouts", DEFAULT_LAYOUT +
-                                                self.LAYOUT_FILE_EXTENSION))
+        gskey = self.layout_key
+        return self.find_layout_filename(gskey.value, gskey.key,
+                                     self.LAYOUT_FILE_EXTENSION,
+                                     os.path.join(self.install_dir,
+                                                  "layouts", DEFAULT_LAYOUT +
+                                                  self.LAYOUT_FILE_EXTENSION))
+
     def set_layout_filename(self, filename):
         if filename and os.path.exists(filename):
             self.layout = filename
@@ -505,6 +545,23 @@ class Config(ConfigObject):
     layout_filename = property(get_layout_filename, set_layout_filename)
 
 
+    def get_fallback_layout_filename(self):
+        """ Layout file to fallback to, when the initial layout won't load """
+        return self.find_layout_filename(DEFAULT_LAYOUT, "layout",
+                                         self.LAYOUT_FILE_EXTENSION)
+
+    def find_layout_filename(self, filename, description,
+                                    extension = "", final_fallback = ""):
+        """ Find layout file, either the final layout or an import file. """
+        return self._get_user_sys_filename(
+             filename    = filename,
+             description = description,
+             user_filename_func   = lambda x: \
+                 os.path.join(self.user_dir,    "layouts", x) + extension,
+             system_filename_func = lambda x: \
+                 os.path.join(self.install_dir, "layouts", x) + extension,
+             final_fallback       = final_fallback)
+
     # Property theme_filename, linked to gsettings key "theme".
     # theme_filename may only get/set a valid filename,
     # whereas theme also allows to get/set only the basename of a theme.
@@ -512,65 +569,95 @@ class Config(ConfigObject):
         self.theme_notify_add(callback)
 
     def get_theme_filename(self):
-        return self._get_user_sys_filename_gs(
-             gskey                = self.theme_key,
-             user_filename_func   = Theme.build_user_filename,
-             system_filename_func = Theme.build_system_filename,
-             final_fallback       = os.path.join(self.install_dir,
-                                                "themes", DEFAULT_THEME +
-                                                "." + Theme.extension()))
-    def set_theme_filename(self, filename, save = True):
-        if filename and os.path.exists(filename):
-            self.set_theme(filename, save)
+        candidates = self._get_theme_candidates()
+        for theme in candidates:
+            if theme:
+                filename = self._expand_theme_filename(theme)
+                if filename:
+                    return filename
+        return ""
 
-            # remember currently active gtk theme
-            if self.system_theme_tracking_enabled:
-                self.remember_theme(filename)
+    def set_theme_filename(self, filename):
+        if filename and os.path.exists(filename):
+            self.remember_theme(filename)
         else:
             _logger.warning(_format("theme '{filename}' does not exist", \
                                     filename=filename))
 
     theme_filename = property(get_theme_filename, set_theme_filename)
 
+    def _expand_theme_filename(self, theme):
+        """ expand generic theme name """
+        return self._expand_user_sys_filename(theme,
+             user_filename_func   = Theme.build_user_filename,
+             system_filename_func = Theme.build_system_filename)
+
     def remember_theme(self, theme_filename):
-        if self.gdi:   # be defensive
+        if self.system_theme_tracking_enabled and \
+           self.gdi:   # be defensive
             gtk_theme = self.get_gtk_theme()
             theme_assocs = self.system_theme_associations.copy()
             theme_assocs[gtk_theme] = theme_filename
             self.system_theme_associations = theme_assocs
 
-    def apply_theme(self):
-        theme_filename = self.theme_filename
-        _logger.info(_format("Loading theme from '{}'", theme_filename))
+        self.set_theme(theme_filename)
 
-        theme = Theme.load(theme_filename)
+    def apply_theme(self, filename = None):
+        if not filename:
+            filename = self.get_theme_filename()
+
+        _logger.info(_format("Loading theme from '{}'", filename))
+
+        theme = Theme.load(filename)
         if not theme:
-            _logger.error(_format("Unable to read theme '{}'", theme_filename))
+            _logger.error(_format("Unable to read theme '{}'", filename))
         else:
             # Save to gsettings
             # Make sure gsettings is in sync with onboard (LP: 877601)
-            self.theme = theme_filename
+            #self.theme = filename
             theme.apply()
 
             # Fix theme not saved to gesettings when switching
             # system contrast themes.
             # Possible gsettings bug in Precise (wasn't in Oneiric).
-            self.apply()
+            #self.apply()
+
+        return bool(theme)
+
+    def load_theme(self):
+        """
+        Figure out which theme to load and load it.
+        """
+        self.apply_theme()
+
+    def _get_theme_candidates(self):
+        """
+        Return a list of themes to consider for loading.
+        Highest priority first.
+        """
+        candidates = []
 
-    def update_theme_from_system_theme(self):
-        """ Switches themes for system theme tracking """
         if self.system_theme_tracking_enabled:
-            gtk_theme = self.get_gtk_theme()
             theme_assocs = self.system_theme_associations
+            gtk_theme = self.get_gtk_theme()
 
-            new_theme = theme_assocs.get(gtk_theme, None)
-            if not new_theme:
-                new_theme = theme_assocs.get("Default", None)
-                if not new_theme:
-                    new_theme = DEFAULT_THEME
+            candidates += [theme_assocs.get(gtk_theme, ""),
+                           theme_assocs.get("Default", ""),
+                           ""]
+        else:
+            candidates += ["",
+                           "",
+                           self.theme]
 
-            self.theme = new_theme
-            self.apply_theme()
+        if self.system_defaults:
+            theme = self.system_defaults.get("theme", "")
+        else:
+            theme = ""
+        candidates.append(theme)
+
+        candidates.append(DEFAULT_THEME)
+
+        return candidates
 
     def get_gtk_theme(self):
         gtk_settings = Gtk.Settings.get_default()
@@ -592,6 +679,12 @@ class Config(ConfigObject):
              system_filename_func = lambda x: \
                  os.path.join(self.install_dir, "layouts", "images", x))
 
+    def get_user_model_dir(self):
+        return os.path.join(self.user_dir, "models")
+
+    def get_system_model_dir(self):
+        return os.path.join(self.install_dir, "models")
+
     def allow_system_click_type_window(self, allow):
         """ called from hover click button """
         if not self.mousetweaks:
@@ -631,16 +724,33 @@ class Config(ConfigObject):
         return not self.xid_mode and \
                self.auto_show.enabled
 
-    def get_frame_width(self):
-        """ width of the frame around the keyboard """
-        if self.xid_mode:
-            return 1.0
-        elif self.has_window_decoration():
-            return 0.0
-        elif self.window.transparent_background:
-            return 1.0
-        else:
-            return self.UNDECORATED_FRAME_WIDTH
+    def is_force_to_top(self):
+        return self.window.force_to_top or self.is_docking_enabled()
+
+    def is_docking_enabled(self):
+        return self.window.docking_enabled
+
+    def is_dock_expanded(self, orientation_co):
+        return self.window.docking_enabled and orientation_co.dock_expand
+
+    def are_word_suggestions_enabled(self):
+        return self.word_suggestions.enabled and not self.xid_mode
+
+    def are_spelling_suggestions_enabled(self):
+        return self.are_word_suggestions_enabled() and \
+               self.word_suggestions.spelling_suggestions_enabled
+
+    def is_spell_checker_enabled(self):
+        return self.are_spelling_suggestions_enabled() or \
+               self.typing_assistance.auto_capitalization or \
+               self.typing_assistance.auto_correction
+
+    def is_auto_capitalization_enabled(self):
+        return self.typing_assistance.auto_capitalization and not self.xid_mode
+
+    def is_typing_assistance_enabled(self):
+        return self.are_word_suggestions_enabled() or \
+               self.is_auto_capitalization_enabled()
 
     def check_gnome_accessibility(self, parent = None):
         if not self.xid_mode and \
@@ -689,11 +799,11 @@ class Config(ConfigObject):
 
     def has_window_decoration(self):
         """ Force-to-top mode doesn't support window decoration """
-        return self.window.window_decoration and not self.window.force_to_top
+        return self.window.window_decoration and not self.is_force_to_top()
 
     def get_sticky_state(self):
         return not self.xid_mode and \
-               (self.window.window_state_sticky or self.window.force_to_top)
+               (self.window.window_state_sticky or self.is_force_to_top())
 
     def is_inactive_transparency_enabled(self):
         return self.window.enable_inactive_transparency and \
@@ -701,6 +811,9 @@ class Config(ConfigObject):
     def is_keep_aspect_ratio_enabled(self):
         return self.window.keep_aspect_ratio or self.options.keep_aspect_ratio
 
+    def is_mousetweaks_active(self):
+        return self.mousetweaks and self.mousetweaks.is_active()
+
     ####### resize handles #######
     def resize_handles_notify_add(self, callback):
         self.window.resize_handles_notify_add(callback)
@@ -748,6 +861,7 @@ class Config(ConfigObject):
             ids.append(Handle.IDS[handle])
         return " ".join(ids)
 
+                #self.set_drag_handles(config.window.resize_handles)
     ####### Snippets editing #######
     def set_snippet(self, index, value):
         """
@@ -845,11 +959,17 @@ class Config(ConfigObject):
         return result
 
     def _get_user_dir(self):
-        return os.path.join(os.path.expanduser("~"), USER_DIR)
+        return XDGDirs.get_data_home(USER_DIR)
 
+    def get_user_layout_dir(self):
+        return os.path.join(self.user_dir, "layouts/")
 
 class ConfigKeyboard(ConfigObject):
     """Window configuration """
+    DEFAULT_KEY_ACTION = 1 # Release-only, supports long press
+    DEFAULT_KEY_SYNTH  = 0 # XTest
+    DEFAULT_TOUCH_INPUT = TouchInputEnum.MULTI
+    DEFAULT_INPUT_EVENT_SOURCE = InputEventSourceEnum.XINPUT
 
     def _init_keys(self):
         self.schema = SCHEMA_KEYBOARD
@@ -858,10 +978,33 @@ class ConfigKeyboard(ConfigObject):
         self.add_key("show-click-buttons", False)
         self.add_key("sticky-key-release-delay", 0.0)
         self.add_key("sticky-key-behavior", {"all" : "cycle"}, 'a{ss}')
+        self.add_key("long-press-delay", 0.5)
+        self.add_key("default-key-action", self.DEFAULT_KEY_ACTION,
+                                           enum={"single-stroke" : 0,
+                                                 "delayed-stroke" : 1,
+                                                })
+        self.add_key("key-synth", self.DEFAULT_KEY_SYNTH,
+                                           enum={"XTest" : 0,
+                                                 "AT-SPI" : 1,
+                                                })
+        self.add_key("touch-input", self.DEFAULT_TOUCH_INPUT,
+                                           enum={"none" : 0,
+                                                 "single" : 1,
+                                                 "multi" : 2,
+                                                })
+        self.add_key("input-event-source", self.DEFAULT_INPUT_EVENT_SOURCE,
+                                           enum={"GTK" : 0,
+                                                 "XInput" : 1,
+                                                })
+        self.add_key("touch-feedback-enabled", False)
+        self.add_key("touch-feedback-size", 0)
+        self.add_key("audio-feedback-enabled", False)
+        self.add_key("show-secondary-labels", False)
 
 
 class ConfigWindow(ConfigObject):
     """Window configuration """
+    DEFAULT_DOCKING_EDGE = DockingEdge.BOTTOM
 
     def _init_keys(self):
         self.schema = SCHEMA_WINDOW
@@ -878,6 +1021,12 @@ class ConfigWindow(ConfigObject):
         self.add_key("inactive-transparency", 50.0)
         self.add_key("inactive-transparency-delay", 1.0)
         self.add_key("resize-handles", DEFAULT_RESIZE_HANDLES)
+        self.add_key("docking-enabled", False)
+        self.add_key("docking-edge", self.DEFAULT_DOCKING_EDGE,
+                                     enum={"top"    : DockingEdge.TOP,
+                                           "bottom" : DockingEdge.BOTTOM,
+                                          })
+        self.add_key("docking-shrink-workarea", True)
 
         self.landscape = ConfigWindow.Landscape(self)
         self.portrait = ConfigWindow.Portrait(self)
@@ -909,6 +1058,20 @@ class ConfigWindow(ConfigObject):
         self.portrait.width_notify_add(callback)
         self.portrait.height_notify_add(callback)
 
+    def dock_size_notify_add(self, callback):
+        self.landscape.dock_width_notify_add(callback)
+        self.landscape.dock_height_notify_add(callback)
+        self.portrait.dock_width_notify_add(callback)
+        self.portrait.dock_height_notify_add(callback)
+
+    def docking_notify_add(self, callback):
+        self.docking_enabled_notify_add(callback)
+        self.docking_edge_notify_add(callback)
+        self.docking_shrink_workarea_notify_add(callback)
+
+        self.landscape.dock_expand_notify_add(callback)
+        self.portrait.dock_expand_notify_add(callback)
+
     def get_active_opacity(self):
         return 1.0 - self.transparency / 100.0
 
@@ -931,6 +1094,9 @@ class ConfigWindow(ConfigObject):
             self.add_key("y", DEFAULT_Y)
             self.add_key("width", DEFAULT_WIDTH)
             self.add_key("height", DEFAULT_HEIGHT)
+            self.add_key("dock-width", DEFAULT_WIDTH)
+            self.add_key("dock-height", DEFAULT_HEIGHT)
+            self.add_key("dock-expand", True)
 
     class Portrait(ConfigObject):
         def _init_keys(self):
@@ -941,6 +1107,9 @@ class ConfigWindow(ConfigObject):
             self.add_key("y", DEFAULT_Y)
             self.add_key("width", DEFAULT_WIDTH)
             self.add_key("height", DEFAULT_HEIGHT)
+            self.add_key("dock-width", DEFAULT_WIDTH)
+            self.add_key("dock-height", DEFAULT_HEIGHT)
+            self.add_key("dock-expand", True)
 
 
 class ConfigICP(ConfigObject):
@@ -1053,6 +1222,7 @@ class ConfigTheme(ConfigObject):
         self.add_key("key-style", "flat")
         self.add_key("roundrect-radius", 0.0)
         self.add_key("key-size", 100.0)
+        self.add_key("key-stroke-width", 100.0)
         self.add_key("key-fill-gradient", 0.0)
         self.add_key("key-stroke-gradient", 0.0)
         self.add_key("key-gradient-direction", 0.0)
@@ -1068,6 +1238,8 @@ class ConfigTheme(ConfigObject):
         self.background_gradient_notify_add(callback)
         self.key_style_notify_add(callback)
         self.roundrect_radius_notify_add(callback)
+        self.key_size_notify_add(callback)
+        self.key_stroke_width_notify_add(callback)
         self.key_fill_gradient_notify_add(callback)
         self.key_stroke_gradient_notify_add(callback)
         self.key_gradient_direction_notify_add(callback)
@@ -1123,6 +1295,7 @@ class ConfigLockdown(ConfigObject):
         self.add_key("disable-preferences", False)
         self.add_key("disable-quit", False)
         self.add_key("disable-touch-handles", False)
+        self.add_key("disable-keys", [["CTRL", "LALT", "F[0-9]+"]], 'aas')
 
     def lockdown_notify_add(self, callback):
         self.disable_click_buttons_notify_add(callback)
@@ -1152,6 +1325,7 @@ class ConfigGDI(ConfigObject):
         self.add_key("toolkit-accessibility", False)
         self.add_key("gtk-theme", "", writable=False)  # read-only for safety
 
+
 class ConfigGDA(ConfigObject):
     """ Key to check if a11y keyboard is enabled """
 
@@ -1199,3 +1373,58 @@ class ConfigScanner(ConfigObject):
         self.add_key("device-button-map", self.DEFAULT_DEVICE_BUTTON_MAP, 'a{ii}')
         self.add_key("feedback-flash", self.DEFAULT_FEEDBACK_FLASH)
 
+
+class ConfigTypingAssistance(ConfigObject):
+    """ typing-assistance configuration keys"""
+
+    DEFAULT_BACKEND = 0
+
+    def _init_keys(self):
+        self.schema = SCHEMA_TYPING_ASSISTANCE
+        self.sysdef_section = "typing-assistance"
+
+        self.add_key("active-language", "")
+        self.add_key("recent-languages", [], 'as')
+        self.add_key("max-recent-languages", 5)
+        self.add_key("spell-check-backend", self.DEFAULT_BACKEND,
+                                                     enum={"hunspell" : 0,
+                                                           "aspell"   : 1})
+        self.add_key("auto-capitalization", False)
+        self.add_key("auto-correction", False)
+
+        self.word_suggestions = ConfigWordSuggestions(self)
+        self.children = [self.word_suggestions]
+
+        # shortcuts in the root for convenient access
+        self.get_root().wp = self.word_suggestions
+        self.get_root().word_suggestions = self.word_suggestions
+
+
+class ConfigWordSuggestions(ConfigObject):
+    """ word-suggestions configuration keys"""
+
+    def _init_keys(self):
+        self.schema = SCHEMA_WORD_SUGGESTIONS
+        self.sysdef_section = "word-suggestions"
+
+        self.add_key("enabled", False)
+        self.add_key("auto-learn", True)
+        self.add_key("punctuation-assistance", True)
+        self.add_key("stealth-mode", False)
+        self.add_key("accent-insensitive", True)
+        self.add_key("max-word-choices", 5)
+        self.add_key("spelling-suggestions-enabled", True)
+
+        self.add_key("show-context-line", False)
+
+    def word_prediction_notify_add(self, callback):
+        self.auto_learn_notify_add(callback)
+        self.punctuation_assistance_notify_add(callback)
+        self.stealth_mode_notify_add(callback)
+
+    def can_auto_learn(self):
+        return self.enabled and \
+               self.auto_learn and \
+               not self.stealth_mode
+
+
