@@ -51,7 +51,7 @@ class ConfigObject(object):
         self.gskeys = {}           # key-value objects {property name, GSKey()}
         self.sysdef_section = None # system defaults section name
         self.system_defaults = {}  # system defaults {property name, value}
-        self._osk_dconf = None 
+        self._osk_dconf = None
 
         # add keys in here
         self._init_keys()
@@ -182,7 +182,7 @@ class ConfigObject(object):
                             for callback in callbacks:
                                 callback(value)
 
-                        GObject.idle_add(notify, 
+                        GLib.idle_add(notify,
                                         getattr(self, _NOTIFY_CALLBACKS.format(prop)),
                                         value)
                     else:
@@ -214,8 +214,9 @@ class ConfigObject(object):
                    getattr(self, _CAN_SET_HOOK +_prop)(value):
 
                 if save:
-                    if value != _gskey.value:
+                    if value != _gskey.value or _gskey.modified:
                         self.set_unpacked(_gskey, value)
+                        _gskey.modified = False
 
                 _gskey.value = value
 
@@ -235,7 +236,9 @@ class ConfigObject(object):
         self.init_from_gsettings()
 
         # let system defaults override gsettings
-        if self.use_system_defaults:
+        use_system_defaults = self.use_system_defaults or \
+                              self._check_hints_file()
+        if use_system_defaults:
             self.init_from_system_defaults()
             self.use_system_defaults = False    # write to gsettings
 
@@ -246,6 +249,29 @@ class ConfigObject(object):
                 if not value is None:
                     gskey.value = value
 
+        return use_system_defaults
+
+
+    def _check_hints_file(self):
+        """
+        Use system defaults if this file exists, then delete it.
+        Workaround for the difficult to access gsettings/dconf
+        database when running Onboard in lightdm."
+        """
+        filename = "/tmp/onboard-use-system-defaults"
+        if os.path.exists(filename):
+            _logger.warning("Hint file '{}' exists; applying system defaults." \
+                         .format(filename))
+            try:
+                os.remove(filename)
+            except IOError as ex:
+                errno = ex.errno
+                errstr = os.strerror(errno)
+                _logger.error("failed to remove hint file '{}': {} ({})" \
+                            .format(filename, errstr, errno))
+            return True
+        return False
+
     def init_from_gsettings(self):
         """ init propertiy values from gsettings """
 
@@ -258,8 +284,10 @@ class ConfigObject(object):
     def init_from_system_defaults(self):
         """ fill property values with system defaults """
 
+        self.delay()
         for prop, value in list(self.system_defaults.items()):
             setattr(self, prop, value)  # write to gsettings
+        self.apply()
 
         for child in self.children:
             child.init_from_system_defaults()
@@ -308,7 +336,15 @@ class ConfigObject(object):
                                   path=self.schema,
                                   gskey=gskey.key, value=value))
 
-            setattr(self, gskey.prop, value)
+            try:
+                setattr(self, gskey.prop, value)
+            except Exception as ex:
+                _logger.error("Migrating dconf key failed: '{key}={value}'; "
+                              "possibly due to incompatible dconf type; "
+                              "skipping this key. "
+                              "Exception: {exception}" \
+                              .format(key=dconf_key, value=value,
+                                      exception=unicode_str(ex)))
 
     def migrate_dconf_key(self, dconf_key, key):
         gskey = self.find_key(key)
@@ -376,31 +412,27 @@ class ConfigObject(object):
         """
 
         filepath = filename
-        if filename and not os.path.exists(filename):
+
+        if filename and not os.path.isfile(filename):
             # assume filename is just a basename instead of a full file path
             _logger.debug(_format("{description} '{filename}' not found yet, "
                                   "retrying in default paths", \
                                   description=description, filename=filename))
 
-            if user_filename_func:
-                filepath = user_filename_func(filename)
-                if not os.path.exists(filepath):
-                    filepath = ""
-
-            if  not filepath and system_filename_func:
-                filepath = system_filename_func(filename)
-                if not os.path.exists(filepath):
-                    filepath = ""
-
+            filepath = ConfigObject._expand_user_sys_filename(
+                                                     filename,
+                                                     user_filename_func,
+                                                     system_filename_func)
             if not filepath:
                 _logger.info(_format("unable to locate '{filename}', "
                                      "loading default {description} instead",
                                      description=description,
                                      filename=filename))
+
         if not filepath and not final_fallback is None:
             filepath = final_fallback
 
-        if not os.path.exists(filepath):
+        if not os.path.isfile(filepath):
             _logger.error(_format("failed to find {description} '{filename}'",
                                   description=description, filename=filename))
             filepath = ""
@@ -410,6 +442,24 @@ class ConfigObject(object):
 
         return filepath
 
+    @staticmethod
+    def _expand_user_sys_filename(filename, \
+                                  user_filename_func = None,
+                                  system_filename_func = None):
+        result = filename
+        if result and not os.path.isfile(result):
+            if user_filename_func:
+                result = user_filename_func(filename)
+                if not os.path.isfile(result):
+                    result = ""
+
+            if  not result and system_filename_func:
+                result = system_filename_func(filename)
+                if not os.path.isfile(result):
+                    result = ""
+
+        return result
+
     @staticmethod
     def get_unpacked_string_list(gskey, type_spec):
         """ Store dictionary in a gsettings list key """
@@ -441,7 +491,7 @@ class ConfigObject(object):
         """ very crude hard coded behavior, fixme as needed """
         if type_spec == "a{ss}":
             _dict = ConfigObject._list_to_dict(_list, str, num_values = 1)
-            return dict([key, value[0]] for key, value in _dict.items())  
+            return dict([key, value[0]] for key, value in _dict.items())
 
         if type_spec == "a{s[ss]}":
             return ConfigObject._list_to_dict(_list, str, num_values = 2)
@@ -478,7 +528,10 @@ class ConfigObject(object):
         filename = None
         parser = configparser.SafeConfigParser()
         try:
-            filename = parser.read(paths, "UTF-8")
+            if sys.version_info.major == 2:
+                filename = parser.read(paths)
+            else:
+                filename = parser.read(paths, "UTF-8")
         except configparser.ParsingError as ex:
             _logger.error(_("Failed to read system defaults. " + \
                             unicode_str(ex)))
@@ -531,6 +584,14 @@ class ConfigObject(object):
             _logger.warning(_format("System defaults: Unknown key '{}' "
                                     "in section '{}'",
                                     sysdef, self.sysdef_section))
+        elif gskey.enum:
+            try:
+                value = gskey.enum[value]
+            except KeyError as ex:
+                _logger.warning(_format("System defaults: Invalid enum value"
+                                        " for key '{}' in section '{}':"
+                                        " {}",
+                                        sysdef, self.sysdef_section, ex))
         else:
             _type = type(gskey.default)
             str_type = str if sys.version_info.major >= 3 \
@@ -569,7 +630,8 @@ class GSKey:
         self.enum        = enum        # dict of enum choices {si}
         self.value       = default     # current property value
         self.writable    = writable    # If False, never write the key
-                                       #    to gsettings even on accident.
+                                       #    to gsettings, even on accident.
+        self.modified    = False       # If True, force writing to gsettings
 
     def is_default(self):
         return self.value == self.default
@@ -618,4 +680,3 @@ class GSKey:
             else:
                 self.settings[self.key] = value
 
-
