@@ -9,28 +9,29 @@ _logger = logging.getLogger("OnboardGtk")
 
 import sys
 import time
-import traceback
 import signal
 import os.path
 
-import dbus
-from dbus.mainloop.glib import DBusGMainLoop
+try:
+    import dbus
+    import dbus.service
+    import dbus.mainloop.glib
+except ImportError:
+    pass
 
-from gi.repository import GObject, Gio, Gdk, Gtk, GLib
 
-import virtkey
+from gi.repository import GLib, Gdk, Gtk
 
-from Onboard.Indicator import Indicator
-from Onboard.Keyboard import Keyboard
-from Onboard.Scanner import Scanner
-from Onboard.KeyGtk import *
-from Onboard.KbdWindow import KbdWindow, KbdPlugWindow
-from Onboard.KeyboardGTK import KeyboardGTK
+from Onboard.KbdWindow       import KbdWindow, KbdPlugWindow
+from Onboard.Keyboard        import Keyboard
+from Onboard.KeyboardWidget  import KeyboardWidget
+from Onboard.Indicator       import Indicator
 from Onboard.LayoutLoaderSVG import LayoutLoaderSVG
-from Onboard.Appearance import Theme, ColorScheme
-from Onboard.IconPalette import IconPalette
-from Onboard.utils      import show_confirmation_dialog, CallOnce, Process, \
-                               unicode_str
+from Onboard.Appearance      import ColorScheme
+from Onboard.IconPalette     import IconPalette
+from Onboard.Exceptions      import LayoutFileError
+from Onboard.utils           import show_confirmation_dialog, CallOnce, Process, \
+                                    unicode_str
 import Onboard.osk as osk
 
 ### Config Singleton ###
@@ -43,14 +44,14 @@ import Onboard.KeyCommon
 app = "onboard"
 DEFAULT_FONTSIZE = 10
 
-class OnboardGtk(Gtk.Application):
+
+class OnboardGtk(object):
     """
     Main controller class for Onboard using GTK+
     """
 
-    ONBOARD_APP_ID = "net.launchpad.onboard"
+    DBUS_NAME = "org.onboard.Onboard"
 
-    """ The keyboard widget """
     keyboard = None
 
     def __init__(self):
@@ -61,40 +62,56 @@ class OnboardGtk(Gtk.Application):
         GLib.set_prgname(str(app))
         Gdk.set_program_class(app[0].upper() + app[1:])
 
-        # Use D-bus main loop by default
-        DBusGMainLoop(set_as_default=True)
+        # no python3-dbus on Fedora17
+        if not "dbus" in globals():
+            _logger.warning("D-Bus bindings unavailable. "
+                            "Onboard will start with reduced functionality. "
+                            "Single-instance check, "
+                            "D-Bus service and "
+                            "hover click are disabled.")
+        else:
+            # Use D-bus main loop by default
+            dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
 
-        # Onboard in Ubuntu on first start silently embeds itself into
-        # gnome-screen-saver and stays like this until embedding is manually
-        # turned off.
-        # If gnome's "Typing Assistent" is disabled, only show onboard in
-        # gss when there is already a non-embedded instance running in
-        # the user session (LP: 938302).
-        bus = dbus.SessionBus()
-        has_remote_instance = bus.name_has_owner(self.ONBOARD_APP_ID)
-        if config.xid_mode and \
-           not (config.gnome_a11y and \
-                config.gnome_a11y.screen_keyboard_enabled):
-            if Process.was_launched_by("gnome-screensaver") and \
-               not has_remote_instance:
+            # Yield to GNOME Shell's keyboard before any other D-Bus activity
+            # to reduce the chance for D-Bus timeouts when enabling a11y keboard.
+            if not self._can_show_in_current_desktop():
                 sys.exit(0)
 
-        if config.options.allow_multiple_instances or \
-           config.xid_mode:
-            app_flags = Gio.ApplicationFlags.NON_UNIQUE
-        else:
-            app_flags = Gio.ApplicationFlags.FLAGS_NONE
+            # Check if there is already a Onboard instance running
+            bus = dbus.SessionBus()
+            has_remote_instance = bus.name_has_owner(self.DBUS_NAME)
 
-        super(OnboardGtk, self).__init__(application_id=OnboardGtk.ONBOARD_APP_ID,
-                                         flags=app_flags)
+            # Onboard in Ubuntu on first start silently embeds itself into
+            # gnome-screensaver and stays like this until embedding is manually
+            # turned off.
+            # If gnome's "Typing Assistent" is disabled, only show onboard in
+            # gss when there is already a non-embedded instance running in
+            # the user session (LP: 938302).
+            if config.xid_mode and \
+               not (config.gnome_a11y and \
+                    config.gnome_a11y.screen_keyboard_enabled):
+                if Process.was_launched_by("gnome-screensaver") and \
+                   not has_remote_instance:
+                    sys.exit(0)
+
+            # Embedded instances can't become primary instances
+            if not config.xid_mode:
+                if has_remote_instance and \
+                   not config.options.allow_multiple_instances:
+                    # Present remote instance
+                    remote = bus.get_object(self.DBUS_NAME, ServiceOnboardKeyboard.PATH)
+                    remote.Show(dbus_interface=ServiceOnboardKeyboard.IFACE)
+                    _logger.info("Exiting: Not the primary instance.")
+                    sys.exit(0)
+
+                # Register our dbus name
+                self._bus_name = dbus.service.BusName(self.DBUS_NAME, bus)
+
+        self.init()
 
         _logger.info("Entering mainloop of onboard")
-        self.run(None)
-
-        # Additional instances after the first one don't open main windows.
-        # Make sure that startup is announced as complete or unity will
-        # block the launcher icon for 3 seconds.
-        Gdk.notify_startup_complete()
+        Gtk.main()
 
         # Shut up error messages on SIGTERM in lightdm:
         # "sys.excepthook is missing, lost sys.stderr"
@@ -109,18 +126,6 @@ class OnboardGtk(Gtk.Application):
         except:
             pass
 
-    def do_activate(self):
-        """
-        App instance entry point.
-        This is always called in the context of the first instance.
-        """
-        if len(self.get_windows()) == 0:
-            self.init()
-            self.add_window(self._window)
-        else:
-            if self.keyboard:
-                self.keyboard.set_visible(True)
-
     def init(self):
         self.keyboard_state = None
         self.vk_timer = None
@@ -128,45 +133,51 @@ class OnboardGtk(Gtk.Application):
         self._connections = []
         self._window = None
         self.status_icon = None
+        self.service_keyboard = None
 
         # finish config initialization
         config.init()
 
         # Release pressed keys when onboard is killed.
-        # Don't keep enter key stuck when being killed from lightdm.
+        # Don't keep enter key stuck when being killed by lightdm.
         self._osk_util = osk.Util()
         self._osk_util.set_unix_signal_handler(signal.SIGTERM, self.on_sigterm)
         self._osk_util.set_unix_signal_handler(signal.SIGINT, self.on_sigint)
 
         sys.path.append(os.path.join(config.install_dir, 'scripts'))
 
-        # Create the keyboard
-        # Care for toolkit independency only once there is another
+        # Create the central keyboard model
+        self.keyboard = Keyboard()
+        
+        # Create the initial keyboard widget
+        # Care for toolkit independence only once there is another
         # supported one besides GTK.
-        self.keyboard = KeyboardGTK()
-
-        # load the initial layout
-        _logger.info("Loading initial layout")
-        self.reload_layout()
+        self.keyboard_widget = KeyboardWidget(self.keyboard)
 
         # create the main window
         if config.xid_mode:    # XEmbed mode for gnome-screensaver?
-            self._window = KbdPlugWindow()
+            # no icp, don't flash the icon palette in lightdm
+
+            self._window = KbdPlugWindow(self.keyboard_widget)
 
             # write xid to stdout
             sys.stdout.write('%d\n' % self._window.get_id())
             sys.stdout.flush()
         else:
-            self._window = KbdWindow()
+            icp = IconPalette()
+            icp.set_layout_view(self.keyboard_widget)
+            icp.connect("activated", self._on_icon_palette_acticated)
+
+            self._window = KbdWindow(self.keyboard_widget, icp)
             self.do_connect(self._window, "quit-onboard",
                             lambda x: self.do_quit_onboard())
 
-        if not config.xid_mode:  # don't flash the icon palette in lightdm
-            icp = IconPalette()
-            icp.connect("activated", self._on_icon_palette_acticated)
-            self._window.icp = icp
         self._window.application = self
-        self._window.set_keyboard(self.keyboard)
+        config.main_window = self._window # need this to access screen properties
+
+        # load the initial layout
+        _logger.info("Loading initial layout")
+        self.reload_layout()
 
         # Handle command line options x, y, size after window creation
         # because the rotation code needs the window's screen.
@@ -190,8 +201,13 @@ class OnboardGtk(Gtk.Application):
                 self._window.write_window_rect(orientation, rect)
                 self._window.restore_window_rect() # move/resize early
 
+        # export dbus service
+        if not config.xid_mode and \
+           "dbus" in globals():
+            self.service_keyboard = ServiceOnboardKeyboard(self.keyboard_widget)
+
         # show/hide the window
-        self.keyboard.set_startup_visibility()
+        self.keyboard_widget.set_startup_visibility()
 
         # keep keyboard window and icon palette on top of dash
         if not config.xid_mode: # be defensive, not necessary when embedding
@@ -201,6 +217,7 @@ class OnboardGtk(Gtk.Application):
         # connect notifications for keyboard map and group changes
         self.keymap = Gdk.Keymap.get_default()
         self.do_connect(self.keymap, "keys-changed", self.cb_keys_changed) # map changes
+        self.do_connect(self.keymap, "state-changed", self.cb_state_changed)
         Gdk.event_handler_set(cb_any_event, self)          # group changes
 
         # connect config notifications here to keep config from holding
@@ -208,20 +225,27 @@ class OnboardGtk(Gtk.Application):
         once = CallOnce(50).enqueue  # delay callbacks by 50ms
         reload_layout       = lambda x: once(self.reload_layout_and_present)
         update_ui           = lambda x: once(self._update_ui)
-        redraw              = lambda x: once(self.keyboard.redraw)
-        update_transparency = lambda x: once(self.keyboard.update_transparency)
+        update_ui_no_resize = lambda x: once(self._update_ui_no_resize)
+        update_transparency = lambda x: once(self.keyboard_widget.update_transparency)
         update_inactive_transparency = \
-                              lambda x: once(self.keyboard.update_inactive_transparency)
+                              lambda x: once(self.keyboard_widget.update_inactive_transparency)
 
         # general
         config.auto_show.enabled_notify_add(lambda x: \
-                                    self.keyboard.update_auto_show())
+                                    self.keyboard_widget.update_auto_show())
+
+        # keyboard
+        config.keyboard.key_synth_notify_add(reload_layout)
+        config.keyboard.input_event_source_notify_add(lambda x: \
+                                    self.keyboard.update_input_event_source())
+        config.keyboard.show_secondary_labels_notify_add(update_ui)
+
 
         # window
         config.window.window_state_sticky_notify_add(lambda x: \
                                    self._window.update_sticky_state())
-        config.window.window_decoration_notify_add(self._update_window_options)
-        config.window.force_to_top_notify_add(self._update_window_options)
+        config.window.window_decoration_notify_add(self._on_window_options_changed)
+        config.window.force_to_top_notify_add(self._on_window_options_changed)
         config.window.keep_aspect_ratio_notify_add(update_ui)
 
         config.window.transparency_notify_add(update_transparency)
@@ -229,6 +253,7 @@ class OnboardGtk(Gtk.Application):
         config.window.transparent_background_notify_add(update_ui)
         config.window.enable_inactive_transparency_notify_add(update_transparency)
         config.window.inactive_transparency_notify_add(update_inactive_transparency)
+        config.window.docking_notify_add(self._update_docking)
 
         # layout
         config.layout_filename_notify_add(reload_layout)
@@ -241,29 +266,37 @@ class OnboardGtk(Gtk.Application):
         config.theme_settings.color_scheme_filename_notify_add(reload_layout)
         config.theme_settings.key_label_font_notify_add(reload_layout)
         config.theme_settings.key_label_overrides_notify_add(reload_layout)
-        config.theme_settings.key_size_notify_add(update_ui) # for label size
-        config.theme_settings.theme_attributes_notify_add(redraw)
+        config.theme_settings.theme_attributes_notify_add(update_ui)
 
         # snippets
         config.snippets_notify_add(reload_layout)
 
+        # word suggestions
+        config.word_suggestions.show_context_line_notify_add(update_ui)
+        config.word_suggestions.enabled_notify_add(lambda x: \
+                                 self.keyboard.on_word_suggestions_enabled(x))
+        config.typing_assistance.active_language_notify_add(lambda x: \
+                                 self.keyboard.on_active_lang_id_changed())
+        config.typing_assistance.spell_check_backend_notify_add(lambda x: \
+                                 self.keyboard.on_spell_checker_changed())
+        config.typing_assistance.auto_capitalization_notify_add(lambda x: \
+                                 self.keyboard.on_word_suggestions_enabled(x))
+        config.word_suggestions.spelling_suggestions_enabled_notify_add(lambda x: \
+                                 self.keyboard.on_spell_checker_changed())
+
         # universal access
         config.scanner.enabled_notify_add(self.keyboard._on_scanner_enabled)
-        GObject.idle_add(self.keyboard.enable_scanner, config.scanner.enabled)
 
         config.window.resize_handles_notify_add(lambda x: \
-                                    self.keyboard.update_resize_handles())
+                                    self.keyboard_widget.update_resize_handles())
 
         # misc
         config.keyboard.show_click_buttons_notify_add(update_ui)
         config.lockdown.lockdown_notify_add(update_ui)
-        config.clickmapper.state_notify_add(update_ui)
         if config.mousetweaks:
-            config.mousetweaks.state_notify_add(update_ui)
+            config.mousetweaks.state_notify_add(update_ui_no_resize)
 
         # create status icon
-        # Indicator is a singleton to allow recreating the keyboard
-        # window on changes to the "force_to_top" setting.
         self.status_icon = Indicator()
         self.status_icon.set_keyboard_window(self._window)
         self.do_connect(self.status_icon, "quit-onboard",
@@ -367,7 +400,7 @@ class OnboardGtk(Gtk.Application):
 
     # Method concerning the icon palette
     def _on_icon_palette_acticated(self, widget):
-        self.keyboard.toggle_visible()
+        self.keyboard_widget.toggle_visible()
 
     def cb_icp_in_use_toggled(self, icp_in_use):
         """
@@ -414,28 +447,42 @@ class OnboardGtk(Gtk.Application):
 
         TODO would be nice if appeared to iconify to taskbar
         """
-        self.keyboard.toggle_visible()
+        self.keyboard_widget.toggle_visible()
 
 
     # keyboard layout changes
-    def cb_keys_changed(self, *args):
+    def cb_keys_changed(self, keymap):
         self.reload_layout()
 
+    # modifier changes
+    def cb_state_changed(self, keymap):
+        _logger.debug("keyboard state changed to 0x{:x}" \
+                      .format(keymap.get_modifier_state()))
+        mod_mask = keymap.get_modifier_state()
+        self.keyboard.set_modifiers(mod_mask)
+
     def cb_vk_timer(self):
         """
         Timer callback for polling until virtkey becomes valid.
         """
         if self.get_vk():
             self.reload_layout(force_update=True)
-            GObject.source_remove(self.vk_timer)
+            GLib.source_remove(self.vk_timer)
             self.vk_timer = None
             return False
         return True
 
     def _update_ui(self):
-        if self.keyboard:
-            self.keyboard.update_ui()
-            self.keyboard.redraw()
+        self.keyboard.invalidate_ui()
+        self.keyboard.commit_ui_updates()
+
+    def _update_ui_no_resize(self):
+        self.keyboard.invalidate_ui_no_resize()
+        self.keyboard.commit_ui_updates()
+
+    def _on_window_options_changed(self, value = None):
+        self._update_window_options()
+        self.keyboard.commit_ui_updates()
 
     def _update_window_options(self, value = None):
         window = self._window
@@ -443,21 +490,31 @@ class OnboardGtk(Gtk.Application):
             window.update_window_options()
             if window.icp:
                 window.icp.update_window_options()
-            self._update_ui()
+            self.keyboard.invalidate_ui()
+
+    def _update_docking(self, value = None):
+        self._update_window_options()
+
+        # give WM time to settle, else move to the strut position may fail
+        GLib.idle_add(self._update_docking_delayed)
+
+    def _update_docking_delayed(self):
+        self._window.on_docking_notify()
+        self.keyboard.invalidate_ui()  # show/hide the move button
+#        self.keyboard.commit_ui_updates() # redundant
 
     def on_gtk_theme_changed(self, gtk_theme = None):
         """
         Switch onboard themes in sync with gtk-theme changes.
         """
-        config.update_theme_from_system_theme()
+        config.load_theme()
 
     def on_gtk_font_dpi_changed(self):
         """
         Refresh the key's pango layout objects so that they can adapt
         to the new system dpi setting.
         """
-        if self.keyboard:
-            self.keyboard.refresh_pango_layouts()
+        self.keyboard_widget.refresh_pango_layouts()
         self._update_ui()
 
         return False
@@ -472,8 +529,7 @@ class OnboardGtk(Gtk.Application):
         with active transparency
         """
         self.reload_layout(force_update = True)
-        if self.keyboard:
-            self.keyboard.update_transparency()
+        self.keyboard_widget.update_transparency()
 
     def reload_layout(self, force_update=False):
         """
@@ -488,8 +544,7 @@ class OnboardGtk(Gtk.Application):
                 vk.reload() # reload keyboard names
                 keyboard_state = (vk.get_layout_symbols(),
                                   vk.get_current_group_name())
-            except virtkey.error:
-                #traceback.print_exc(file=sys.stdout)
+            except osk.error:
                 self.reset_vk()
                 force_update = True
                 _logger.warning("Keyboard layout changed, but retrieving "
@@ -497,18 +552,28 @@ class OnboardGtk(Gtk.Application):
 
         if self.keyboard_state != keyboard_state or force_update:
             self.keyboard_state = keyboard_state
-            self.load_layout(config.layout_filename,
-                             config.theme_settings.color_scheme_filename)
+
+            layout_filename = config.layout_filename
+            color_scheme_filename = config.theme_settings.color_scheme_filename
+
+            try:
+                self.load_layout(layout_filename, color_scheme_filename)
+            except LayoutFileError as ex:
+                _logger.error("Layout error: " + unicode_str(ex) + ". Falling back to default layout.")
+
+                # last ditch effort to load the default layout
+                self.load_layout(config.get_fallback_layout_filename(),
+                                 color_scheme_filename)
 
         # if there is no X keyboard, poll until it appears (if ever)
         if not vk and not self.vk_timer:
-            self.vk_timer = GObject.timeout_add_seconds(1, self.cb_vk_timer)
+            self.vk_timer = GLib.timeout_add_seconds(1, self.cb_vk_timer)
 
     def load_layout(self, layout_filename, color_scheme_filename):
         _logger.info("Loading keyboard layout from " + layout_filename)
         if (color_scheme_filename):
             _logger.info("Loading color scheme from " + color_scheme_filename)
-        
+
         vk = self.get_vk()
 
         color_scheme = ColorScheme.load(color_scheme_filename) \
@@ -516,12 +581,11 @@ class OnboardGtk(Gtk.Application):
         layout = LayoutLoaderSVG().load(vk, layout_filename, color_scheme)
 
         self.keyboard.cleanup()
-        self.keyboard.vk = vk
+        self.keyboard.init_key_synth(vk)
         self.keyboard.layout = layout
         self.keyboard.color_scheme = color_scheme
-        self.keyboard.initial_update()
-        self.keyboard.update_ui()
-        self.keyboard.redraw()
+        self.keyboard.on_layout_loaded()
+        self.keyboard_widget.on_layout_loaded()
 
         if self._window and self._window.icp:
             self._window.icp.queue_draw()
@@ -530,9 +594,9 @@ class OnboardGtk(Gtk.Application):
         if not self._vk:
             try:
                 # may fail if there is no X keyboard (LP: 526791)
-                self._vk = virtkey.virtkey()
+                self._vk = osk.Virtkey()
 
-            except virtkey.error as e:
+            except osk.error as e:
                 t = time.time()
                 if t > self._vk_error_time + .2: # rate limit to once per 200ms
                     _logger.warning("vk: " + unicode_str(e))
@@ -567,26 +631,132 @@ class OnboardGtk(Gtk.Application):
 
         self.status_icon.set_keyboard_window(None)
         self._window.cleanup()
-        # Stops the GTK main loop
         self._window.destroy()
         self._window = None
+        Gtk.main_quit()
 
     def final_cleanup(self):
         config.final_cleanup()
 
+    @staticmethod
+    def _can_show_in_current_desktop():
+        """
+        When GNOME's "Typing Assistent" is enabled in GNOME Shell, Onboard 
+        starts simultaneously with the Shell's built-in screen keyboard. 
+        With GNOME Shell 3.5.4-0ubuntu2 there is no known way to choose
+        one over the other (LP: 879942). 
+
+        Adding NotShowIn=GNOME; to onboard-autostart.desktop prevents it 
+        from running not only in GNOME Shell, but also in the GMOME Fallback 
+        session, which is undesirable. Both share the same xdg-desktop name.
+
+        -> Do it ourselves: optionally check for GNOME Shell and yield to the
+        built-in keyboard.
+        """
+        result = True
+
+        if config.options.not_show_in:
+            bus = dbus.SessionBus()
+            current = os.environ.get("XDG_CURRENT_DESKTOP", "")
+            names = config.options.not_show_in.split(",")
+            for name in names:                
+                if name == "GNOME":
+                    if bus.name_has_owner("org.gnome.Shell"):
+                        result = False
+                elif name == current:
+                    result = False
+
+            if not result:
+                _logger.info("Command line option not-show-in={} forbids running in "
+                             "current desktop environment '{}'; exiting." \
+                             .format(names, current))
+        return result
+
+
+if "dbus" in globals():
+    class ServiceOnboardKeyboard(dbus.service.Object):
+        """
+        Onboard's D-Bus service.
+        """
+
+        PATH = "/org/onboard/Onboard/Keyboard"
+        IFACE = "org.onboard.Onboard.Keyboard"
+
+        class ServiceOnboardException(dbus.DBusException):
+            _dbus_error_name = 'org.onboard.Exception'
+
+
+        def __init__(self, keyboard):
+            super(ServiceOnboardKeyboard, self).__init__(dbus.SessionBus(), self.PATH)
+            self._keyboard = keyboard
+
+        @dbus.service.method(dbus_interface=IFACE)
+        def Show(self):
+            self._keyboard.set_visible(True)
+
+        @dbus.service.method(dbus_interface=IFACE)
+        def Hide(self):
+            self._keyboard.set_visible(False)
+
+        @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
+                             in_signature='ss', out_signature='v')
+        def Get(self, iface, prop):
+            if iface == self.IFACE:
+                if prop == 'Visible':
+                    return self._keyboard.is_visible()
+                else:
+                    raise self.ServiceOnboardException(\
+                        ('Unknown property \'{0}\'').format(prop))
+            else:
+                raise self.ServiceOnboardException(\
+                    ('Unknown interface \'{0}\'').format(iface))
+
+        @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE, in_signature='ssv')
+        def Set(self, iface, prop, value):
+            if iface == self.IFACE:
+                if prop == 'Visible':
+                    raise self.ServiceOnboardException(\
+                        ('Property \'{0}\' is read-only').format(prop))
+                else:
+                    raise self.ServiceOnboardException(\
+                        ('Unknown property \'{0}\'').format(prop))
+            else:
+                raise self.ServiceOnboardException(\
+                    ('Unknown interface \'{0}\'').format(iface))
+
+        @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
+                             in_signature='s', out_signature='a{sv}')
+        def GetAll(self, iface):
+            if iface == self.IFACE:
+                return { 'Visible': self._keyboard.is_visible() }
+            else:
+                raise self.ServiceOnboardException(\
+                    ('Unknown interface \'{0}\'').format(iface))
+
+        @dbus.service.method(dbus_interface=dbus.INTROSPECTABLE_IFACE, out_signature='s')
+        def Introspect(self):
+            ref = dbus.service.Object.Introspect(self, self._object_path, self.connection)
+
+            iface = '  <interface name="{}">\n' \
+                    '      <property name="Visible" type="b" access="read"/>\n' \
+                    '  </interface>\n' \
+                    .format(self.IFACE)
+
+            return ref[:-8] + iface + '</node>\n'
+
+        @dbus.service.signal(dbus_interface=dbus.PROPERTIES_IFACE, signature='sa{sv}as')
+        def PropertiesChanged(self, iface, changed, invalidated):
+            return iface, changed, invalidated
+
 
 def cb_any_event(event, onboard):
-    # Update layout on keyboard group changes
-    # XkbStateNotify maps to Gdk.EventType.NOTHING
-    # https://bugzilla.gnome.org/show_bug.cgi?id=156948
 
-    # Hide bug in Oneirics GTK3
+    # Hide bug in Oneiric's GTK3
     # Suppress ValueError: invalid enum value: 4294967295
-    type = None
     try:
         type = event.type
     except ValueError:
-        pass
+        type = None
 
     if 0: # debug
         a = [event, event.type]
@@ -600,9 +770,15 @@ def cb_any_event(event, onboard):
             a += [event.window, "0x{:x}".format(event.window.get_xid())]
         print(*a)
 
+    # Update layout on keyboard group changes
+    # XkbStateNotify maps to Gdk.EventType.NOTHING
+    # https://bugzilla.gnome.org/show_bug.cgi?id=156948
     if type == Gdk.EventType.NOTHING:
         onboard.reload_layout()
 
+    # Update the cached pango layout object here or Onboard
+    # doesn't get those settings, i.e. label fonts sizes are off
+    # when font dpi changes.
     elif type == Gdk.EventType.SETTING:
         if event.setting.name == "gtk-theme-name":
             onboard.on_gtk_theme_changed()
@@ -610,11 +786,9 @@ def cb_any_event(event, onboard):
                                     "gtk-xft-antialias"
                                     "gtk-xft-hinting",
                                     "gtk-xft-hintstyle"]:
-            # Update the cached pango layout object here or Onboard
-            # doesn't get those settings, in particular the font dpi.
             # For some reason the font sizes are still off when running
             # this immediately. Delay it a little.
-            GObject.idle_add(onboard.on_gtk_font_dpi_changed)
+            GLib.idle_add(onboard.on_gtk_font_dpi_changed)
 
     Gtk.main_do_event(event)
 
