@@ -6,7 +6,7 @@ from __future__ import division, print_function, unicode_literals
 import time
 from math import sqrt, pi
 
-from gi.repository import GObject, Gtk, Gdk
+from gi.repository import GLib, Gtk, Gdk
 
 from Onboard.utils import Rect, Timer
 
@@ -49,6 +49,20 @@ Handle.RESIZERS = (Handle.EAST,
                    Handle.NORTH,
                    Handle.NORTH_EAST)
 
+Handle.TOP_RESIZERS = (
+                   Handle.EAST,
+                   Handle.WEST,
+                   Handle.NORTH_WEST,
+                   Handle.NORTH,
+                   Handle.NORTH_EAST)
+
+Handle.BOTTOM_RESIZERS = (
+                   Handle.EAST,
+                   Handle.SOUTH_EAST,
+                   Handle.SOUTH,
+                   Handle.SOUTH_WEST,
+                   Handle.WEST)
+
 Handle.ALL = Handle.RESIZERS + (Handle.MOVE, )
 
 Handle.CURSOR_TYPES = {
@@ -85,6 +99,11 @@ Handle.RIDS = {
     "M"  : Handle.MOVE}
 
 
+class DockingEdge:
+    TOP = 0
+    BOTTOM = 3
+
+
 class WindowManipulator(object):
     """
     Adds resize and move capability to windows.
@@ -102,30 +121,43 @@ class WindowManipulator(object):
         - Always use threshold when trying to move, otherwise
           clicking to unhide the keyboard window won't work.
     """
-    hit_frame_width = 10           # size of resize corners and edges
-    drag_protection = True         # enable protection threshold
-    temporary_unlock_delay = 6.0   # seconds until protection threshold returns
-                                   #  counts from drag end in fallback mode
-                                   #  counts from drag start in system mode
-                                   #  (unfortunately)
-    min_window_size = (20, 20)
-    _temporary_unlock_time = None
-
-    _drag_start_pointer = None
-    _drag_start_offset  = None
-    _drag_start_rect    = None
-    _drag_handle        = None
-    _drag_active        = False  # has window move/resize actually started yet?
-    _drag_threshold     = 8
-    _drag_snap_threshold = 16
-
-
     def __init__(self):
+        self.hit_frame_width = 10         # size of resize corners and edges
+        self.drag_protection = True       # enable protection threshold
+        self._temporary_unlock_time = None
+
+        # seconds until protection threshold returns
+        # - counts from drag end in fallback mode
+        # - counts from drag start in system mode
+        #   (unfortunately)
+        self.temporary_unlock_delay = 6.0 
+
+        self.min_window_size = (50, 50)
+
+        self._drag_start_pointer = None
+        self._drag_start_offset  = None
+        self._drag_start_rect    = None
+        self._drag_handle        = None
         self._drag_handles = Handle.RESIZERS
+        self._drag_active        = False  # has window move/resize actually started yet?
+        self._drag_threshold     = 8
+        self._drag_snap_threshold = 16
+
+        self._lock_x_axis         = False
+        self._lock_y_axis         = False
+
+        self._last_drag_handle    = None
+        self._monitor_rects       = None  # cache them to save the lookup time 
 
     def set_min_window_size(self, w, h):
         self.min_window_size = (w, h)
 
+    def get_min_window_size(self):
+        return self.min_window_size
+
+    def get_hit_frame_width(self):
+        return self.hit_frame_width
+
     def enable_drag_protection(self, enable):
         self.drag_protection = enable
 
@@ -137,6 +169,9 @@ class WindowManipulator(object):
                     self.get_allocated_width(),
                     self.get_allocated_height())
 
+    def get_drag_start_rect(self):
+        return self._drag_start_rect
+
     def get_drag_window(self):
         return self
 
@@ -153,36 +188,41 @@ class WindowManipulator(object):
         """ Rectangle in canvas coordinates that must not leave the screen. """
         return None
 
-    def handle_press(self, event, move_on_background = False):
-        point = (event.x, event.y)
-        root_point = (event.x_root, event.y_root)
+    def lock_x_axis(self, lock):
+        """ Set to False to constraint movement in x. """
+        self._lock_x_axis = lock
 
-        hit = self.hit_test_move_resize(point)
+    def lock_y_axis(self, lock):
+        """ Set to True to constraint movement in y. """
+        self._lock_y_axis = lock
+
+    def handle_press(self, sequence, move_on_background = False):
+        hit = self.hit_test_move_resize(sequence.point)
         if not hit is None:
             if hit == Handle.MOVE:
-                self.start_move_window(root_point)
+                self.start_move_window(sequence.root_point)
             else:
-                self.start_resize_window(hit, root_point)
+                self.start_resize_window(hit, sequence.root_point)
             return True
 
         if move_on_background:
-            self.start_move_window(root_point)
+            self.start_move_window(sequence.root_point)
             return True
 
         return False
 
-    def handle_motion(self, event, fallback = False):
+    def handle_motion(self, sequence, fallback = False):
         if not self.is_drag_initiated():
             return
 
         snap_to_cursor = False
-        dx = event.x_root - self._drag_start_pointer[0]
-        dy = event.y_root - self._drag_start_pointer[1]
+        x_root, y_root = sequence.root_point
+        dx = x_root - self._drag_start_pointer[0]
+        dy = y_root - self._drag_start_pointer[1]
 
         # distance threshold, protection from accidental drags
         if not self._drag_active:
             d = sqrt(dx*dx + dy*dy)
-
             drag_active = not self.drag_protection
 
             if self.drag_protection:
@@ -213,9 +253,12 @@ class WindowManipulator(object):
             if fallback:
                 self._handle_motion_fallback(dx, dy)
             else:
-                self._handle_motion_system(dx, dy, snap_to_cursor, event)
+                self._handle_motion_system(dx, dy, snap_to_cursor, sequence)
 
-    def _handle_motion_system(self, dx, dy, snap_to_cursor, event):
+            # give keyboard window a chance to react
+            self.on_drag_activated()
+
+    def _handle_motion_system(self, dx, dy, snap_to_cursor, sequence):
         """
         Let the window manager do the moving
         This fixes issues like not reaching edges at high move speed
@@ -225,12 +268,12 @@ class WindowManipulator(object):
         """
         window = self.get_drag_window()
         if window:
-            x = event.x_root
-            y = event.y_root
+            x, y = sequence.root_point
             if self.is_moving():
                 if snap_to_cursor:
-                    x, y = x - dx, y - dy # snap to cursor
-                window.begin_move_drag(1, x, y, event.time)
+                    x = x - dx # snap to cursor
+                    y = y - dy
+                window.begin_move_drag(1, x, y, sequence.time)
             elif self.is_resizing():
 
                 # Compensate for weird begin_resize_drag behaviour:
@@ -240,7 +283,7 @@ class WindowManipulator(object):
                         x, y = x + dx, y + dy
 
                 window.begin_resize_drag(self._drag_handle, 1,
-                                         x, y, event.time)
+                                         x, y, sequence.time)
 
     def stop_system_drag(self):
         """
@@ -258,12 +301,18 @@ class WindowManipulator(object):
         wy = self._drag_start_pointer[1] + dy - self._drag_start_offset[1]
 
         if self._drag_handle == Handle.MOVE:
+            # contrain axis movement
+            if self._lock_x_axis:
+                wx = self.get_drag_window().get_position()[0]
+            if self._lock_y_axis:
+                wx = self.get_drag_window().get_position()[1]
+
             # move window
             x, y = self.limit_position(wx, wy)
             w, h = None, None
         else:
             # resize window
-            wmin, hmin = self.min_window_size  # minimum window size
+            wmin, hmin = self.get_min_window_size()
             rect = self._drag_start_rect
             x0, y0, x1, y1 = rect.to_extents()
             w, h = rect.get_size()
@@ -283,7 +332,7 @@ class WindowManipulator(object):
             if self._drag_handle in [Handle.SOUTH,
                                      Handle.SOUTH_WEST,
                                      Handle.SOUTH_EAST]:
-                y1 = max(wy + h, y0 + wmin)
+                y1 = max(wy + h, y0 + hmin)
 
             x, y, w, h = x0, y0, x1 -x0, y1 - y0
 
@@ -308,6 +357,15 @@ class WindowManipulator(object):
         else:
             window.set_cursor(None)
 
+    def reset_drag_cursor(self):
+        """ set the mouse cursor """
+        window = self.get_window()
+        if not window:
+            return
+
+        if not self._drag_handle is None:    # not dragging a handle?
+            window.set_cursor(None)
+
     def get_drag_cursor_at(self, point):
         hit = self._drag_handle
         if hit is None:
@@ -320,6 +378,7 @@ class WindowManipulator(object):
     def start_move_window(self, point = None):
         self.start_drag(point)
         self._drag_handle = Handle.MOVE
+        self._last_drag_handle = self._drag_handle
 
     def stop_move_window(self):
         self.stop_drag()
@@ -327,8 +386,10 @@ class WindowManipulator(object):
     def start_resize_window(self, handle, point = None):
         self.start_drag(point)
         self._drag_handle = handle
+        self._last_drag_handle = self._drag_handle
 
     def start_drag(self, point = None):
+        self._monitor_rects = None
 
         # Find the pointer position for the occasions when we are
         # not being called from an event (move button).
@@ -382,8 +443,13 @@ class WindowManipulator(object):
 
     def on_drag_initiated(self):
         """
-        User controlled drag has begun.
-        overload this in derived classes.
+        User controlled drag initiated, but drag hasn't actually begun yet.
+        """
+        pass
+
+    def on_drag_activated(self):
+        """
+        Moving/resizing has begun.
         """
         pass
 
@@ -405,6 +471,9 @@ class WindowManipulator(object):
     def is_moving(self):
         return self.is_drag_initiated() and self._drag_handle == Handle.MOVE
 
+    def was_moving(self):
+        return self._last_drag_handle == Handle.MOVE
+
     def is_resizing(self):
         return self.is_drag_initiated() and self._drag_handle  != Handle.MOVE
 
@@ -423,7 +492,7 @@ class WindowManipulator(object):
     def force_into_view(self):
         self.move_into_view()
         if False:  # Only for system drag, not needed when using fallback mode
-            GObject.idle_add(self._do_force_into_view)
+            GLib.idle_add(self._do_force_into_view)
 
     def _do_force_into_view(self):
         """ Works mostly, but occasionally the window disappears... """
@@ -435,24 +504,6 @@ class WindowManipulator(object):
             self._move_resize(_x, _y)
             window.show()
 
-    def get_limit_rects(self):
-        """
-        Screen limits, one rect per monitor. Monitors may have
-        different sizes and arbitrary relative positions.
-        """
-        rects = []
-        screen = self.get_screen()
-        if screen:
-            for i in range(screen.get_n_monitors()):
-                r = screen.get_monitor_geometry(i)
-                rects.append(Rect(r.x, r.y, r.width, r.height))
-        else:
-            rootwin = Gdk.get_default_root_window()
-            r = Rect.from_position_size(rootwin.get_position(),
-                                    (rootwin.get_width(), rootwin.get_height()))
-            rects.append(r)
-        return rects
-
     def limit_size(self, rect):
         """
         Limits the given window rect to fit on screen.
@@ -466,56 +517,30 @@ class WindowManipulator(object):
             r.h = limits.h - 20
         return r
 
-    def limit_position(self, x, y, visible_rect = None):
+    def limit_position(self, x, y, visible_rect = None, limit_rects = None):
         """
         Limits the given window position to keep the current
         always_visible_rect fully in view.
         """
         # rect to stay always visible, in canvas coordinates
-        r = visible_rect
-        if r is None:
-            r = self.get_always_visible_rect()
+        if visible_rect is None:
+            visible_rect = self.get_always_visible_rect()
 
-        if not r is None:
-            r = r.int() # avoid rounding errors
-
-            window = self.get_drag_window()
-            if window:
-                # transform always visible rect to screen coordinates,
-                # take window decoration into account.
-                rs = r.copy()
-                rs.x += x
-                rs.y += y
-
-                dmin = None
-                rsmin = None
-                for limits in self.get_limit_rects():
-                    # get limited candidate rect
-                    rsc = rs.copy()
-                    rsc.x = max(rsc.x, limits.left())
-                    rsc.x = min(rsc.x, limits.right() - rsc.w)
-                    rsc.y = max(rsc.y, limits.top())
-                    rsc.y = min(rsc.y, limits.bottom() - rsc.h)
-
-                    # closest candidate rect wins
-                    cx, cy = rsc.get_center()
-                    dx, dy = rs.x - rsc.x, rs.y - rsc.y
-                    d = dx * dx + dy * dy
-                    if dmin is None or d < dmin:
-                        dmin = d
-                        rsmin = rsc
-
-                x = rsmin.x - r.x
-                y = rsmin.y - r.y
+        if not limit_rects:
+            if not self._monitor_rects:
+                self._monitor_rects = get_monitor_rects(self.get_screen())
+            limit_rects = self._monitor_rects
 
+        x, y = limit_window_position(x, y, visible_rect, limit_rects)
         return x, y
 
     def hit_test_move_resize(self, point):
         canvas_rect = self.get_resize_frame_rect()
         handles = self.get_drag_handles()
+        hit_frame_width = self.get_hit_frame_width()
 
-        w = min(canvas_rect.w / 2, self.hit_frame_width)
-        h = min(canvas_rect.h / 2, self.hit_frame_width)
+        w = min(canvas_rect.w / 2, hit_frame_width)
+        h = min(canvas_rect.h / 2, hit_frame_width)
 
         x, y = point
         x0, y0, x1, y1 = canvas_rect.to_extents()
@@ -560,8 +585,9 @@ class WindowManipulator(object):
         return None
 
     def _move_resize(self, x, y, w = None, h = None):
+        #print("_move_resize", x, y, w, h)
         window = self.get_drag_window()
-        _win = window.get_window()
+        gdk_win = window.get_window()
         if w is None:
             # Stop inserting edge move for now. In unity, when
             # jamming onboard into the lower left corner the keyboard
@@ -573,7 +599,7 @@ class WindowManipulator(object):
             if hasattr(window, "move_resize"):
                 window.move_resize(x, y, w, h) # keyboard window
             else:
-                window.get_window().move_resize(x, y, w, h) # icon palette
+                gdk_win.move_resize(x, y, w, h) # icon palette
 
 
     def _insert_edge_move(self, window, x, y):
@@ -613,6 +639,7 @@ class Orientation:
     class LANDSCAPE: pass
     class PORTRAIT: pass
 
+
 class WindowRectTracker:
     """
     Keeps track of the window rectangle when moving/resizing.
@@ -624,28 +651,21 @@ class WindowRectTracker:
     def __init__(self):
         self._window_rect = None
         self._origin = None
+        self._client_offset = (0, 0)
         self._screen_orientation = None
         self._save_position_timer = Timer()
 
         # init detection of screen "rotation"
         screen = self.get_screen()
-        screen.connect('size-changed', self._on_screen_size_changed)
+        screen.connect('size-changed', self.on_screen_size_changed)
 
     def cleanup(self):
         self._save_position_timer.finish()
 
     def move(self, x, y):
-        """
-        Overload Gtk.Window.move to reliably keep track of
-        the window position.
-        """
         Gtk.Window.move(self, x, y)
 
     def resize(self, w, h):
-        """
-        Overload Gtk.Window.size to reliably keep track of
-        the window size.
-        """
         Gtk.Window.resize(self, w, h)
 
     def move_resize(self, x, y, w, h):
@@ -667,13 +687,19 @@ class WindowRectTracker:
 
     def get_origin(self):
         if self._origin is None:
-            origin = self.get_window().get_origin()
-            if len(origin) == 3:   # What is the first parameter for? Gdk bug?
-                origin = origin[1:]
-            return origin
+            win = self.get_window()
+            if win:
+                origin = win.get_origin()
+                if len(origin) == 3:   # What is the first parameter for? Gdk bug?
+                    origin = origin[1:]
+                return origin
+            return 0
         else:
             return self._origin
 
+    def get_client_offset(self):
+        return self._client_offset
+
     def get_rect(self):
         return self._window_rect
 
@@ -681,18 +707,23 @@ class WindowRectTracker:
         """ This is overloaded in KbdWindow """
         return Gtk.Window.get_visible(self)
 
-    def _on_screen_size_changed(self, screen):
+    def on_screen_size_changed(self, screen):
         """ detect screen rotation (tablets)"""
 
         # Give the screen time to settle, the window manager
-        # may block the move to previously invalid positions.
-        Timer(0.3, self.restore_window_rect)
+        # may block the move to previously invalid positions and
+        # when docked, the slide animation may be drowned out by all
+        # the action in other processes.
+        Timer(1.5, self.on_screen_size_changed_delayed, screen)
+
+    def on_screen_size_changed_delayed(self, screen):
+        self.restore_window_rect()
 
     def get_screen_orientation(self):
         """
         Current orientation of the screen (tablet rotation).
         Only the aspect ratio is taken into account at this time.
-        This appears to cover more cases than loocking at monitor rotation,
+        This appears to cover more cases than looking at monitor rotation,
         in particular with multi-monitor screens.
         """
         screen = self.get_screen()
@@ -708,13 +739,15 @@ class WindowRectTracker:
         """
         visible = self.is_visible()
         if visible:
-            self._window_rect = Rect.from_position_size(Gtk.Window.get_position(self),
-                                                        Gtk.Window.get_size(self))
+            pos  = Gtk.Window.get_position(self)
+            size = Gtk.Window.get_size(self)
             origin      = self.get_window().get_origin()
             if len(origin) == 3:   # What is the first parameter for? Gdk bug?
                 origin = origin[1:]
-            self._origin = origin
 
+            self._window_rect = Rect.from_position_size(pos, size)
+            self._origin = origin
+            self._client_offset = (origin[0] - pos[0], origin[1] - pos[1]) 
             self._screen_orientation = self.get_screen_orientation()
 
     def restore_window_rect(self, startup = False):
@@ -736,11 +769,12 @@ class WindowRectTracker:
         # Give the derived class a chance to modify the rect,
         # for example to correct the position for auto-show.
         rect = self.on_restore_window_rect(rect)
+        self._window_rect = rect
 
         # move/resize the window
         if startup:
             # gnome-shell doesn't take kindly to an initial move_resize().
-            # The window ends up at (0, 0) on startup and goes back there
+            # The window ends up at (0, 0) on and goes back there
             # repeatedly when hiding and unhiding.
             self.set_default_size(rect.w, rect.h)
             self.move(rect.x, rect.y)
@@ -893,3 +927,124 @@ class DwellProgress(object):
             context.stroke()
 
 
+def limit_window_position(x, y, always_visible_rect, limit_rects = None):
+    """
+    Limits the given window position to keep the
+    always_visible_rect fully in view.
+    """
+    # rect to stay always visible, in canvas coordinates
+    r = always_visible_rect
+
+    if not r is None:
+        r = r.int() # avoid rounding errors
+
+        # transform always visible rect to screen coordinates,
+        # take window decoration into account.
+        rs = r.copy()
+        rs.x += x
+        rs.y += y
+
+        dmin = None
+        rsmin = None
+        for limits in limit_rects:
+            # get limited candidate rect
+            rsc = rs.copy()
+            rsc.x = max(rsc.x, limits.left())
+            rsc.x = min(rsc.x, limits.right() - rsc.w)
+            rsc.y = max(rsc.y, limits.top())
+            rsc.y = min(rsc.y, limits.bottom() - rsc.h)
+
+            # closest candidate rect wins
+            cx, cy = rsc.get_center()
+            dx, dy = rs.x - rsc.x, rs.y - rsc.y
+            d = dx * dx + dy * dy
+            if dmin is None or d < dmin:
+                dmin = d
+                rsmin = rsc
+
+        x = rsmin.x - r.x
+        y = rsmin.y - r.y
+
+    return x, y
+
+def get_monitor_rects(screen):
+    """
+    Screen limits, one rect per monitor. Monitors may have
+    different sizes and arbitrary relative positions.
+    """
+    rects = []
+    if screen:
+        for i in range(screen.get_n_monitors()):
+            r = screen.get_monitor_geometry(i)
+            rects.append(Rect(r.x, r.y, r.width, r.height))
+    else:
+        rootwin = Gdk.get_default_root_window()
+        r = Rect.from_position_size(rootwin.get_position(),
+                                (rootwin.get_width(), rootwin.get_height()))
+        rects.append(r)
+    return rects
+
+def canvas_to_root_window_rect(window, rect):
+    """
+    Convert rect in canvas coordinates to root window coordinates.
+    """
+    gdk_win = window.get_window()
+    if gdk_win:
+        x0, y0 = gdk_win.get_root_coords(rect.x, rect.y)
+        x1, y1 = gdk_win.get_root_coords(rect.x + rect.w,
+                                         rect.y + rect.h)
+        rect = Rect.from_extents(x0, y0, x1, y1)
+    else:
+        rect = Rect()
+
+    return rect
+
+def canvas_to_root_window_point(window, point):
+    """
+    Convert point in canvas coordinates to root window coordinates.
+    """
+    gdk_win = window.get_window()
+    if gdk_win:
+        point = gdk_win.get_root_coords(*point)
+    else:
+        point (0, 0)
+    return point
+
+def get_monitor_dimensions(window):
+    """ Geometry and physical size of the monitor at window. """
+    gdk_win = window.get_window()
+    screen = window.get_screen()
+    if gdk_win and screen:
+        monitor = screen.get_monitor_at_window(gdk_win)
+        r = screen.get_monitor_geometry(monitor)
+        size = (r.width, r.height)
+        size_mm = (screen.get_monitor_width_mm(monitor),
+                   screen.get_monitor_height_mm(monitor))
+
+        # Nexus7 simulation
+        device = None       # keep this at None
+        #device = 1
+        if device == 0:     # dimension unavailable
+            size_mm = 0, 0
+        elif device == 1:     # Nexus 7, as it should report
+            size = 1280, 800
+            size_mm = 150, 94
+
+        return size, size_mm
+    else:
+        return None, None
+
+def physical_to_monitor_pixel_size(window, size_mm, fallback_size = (0, 0)):
+    """ 
+    Convert a physical size in mm to pixels of windows's monitor,
+    """
+    sz, sz_mm = get_monitor_dimensions(window)
+    if sz and sz_mm:
+        w = sz[0] * size_mm[0] / sz_mm[0] \
+            if sz_mm[0] else fallback_size[0]
+        h = sz[1] * size_mm[1] / sz_mm[1] \
+            if sz_mm[0] else fallback_size[1]
+    else:
+        w = h = 0
+    return w, h
+
