@@ -12,7 +12,10 @@ from subprocess import Popen
 from xml.parsers.expat import ExpatError
 from xml.dom import minidom
 import gettext
-from dbus.mainloop.glib import DBusGMainLoop
+try:
+    import dbus.mainloop.glib
+except ImportError:
+    pass
 
 from gi.repository import GObject, Pango, Gdk, Gtk
 
@@ -21,12 +24,11 @@ from Onboard.LayoutLoaderSVG import LayoutLoaderSVG
 from Onboard.SnippetView     import SnippetView
 from Onboard.Appearance      import Theme, ColorScheme
 from Onboard.Scanner         import ScanMode, ScanDevice
+from Onboard.XInput          import XIDeviceManager, XIEventType, XIEventMask
 from Onboard.utils           import show_ask_string_dialog, \
                                     show_confirmation_dialog, \
-                                    unicode_str
-
-from virtkey import virtkey
-from Onboard.osk import Devices
+                                    exists_in_path, \
+                                    unicode_str, open_utf8
 
 app = "onboard"
 
@@ -53,20 +55,116 @@ def format_list_item(text, issystem):
     return text
 
 
-class Settings:
+class DialogBuilder(object):
+    """
+    Utility class for simplified widget setup.
+    Has helpers for connecting widgets to ConfigObject properties, i.e.
+    indirectly to gsettings keys.
+    Mostly borrowed from Gerd Kohlberger's ScannerDialog.
+    """
+
+    def __init__(self, builder):
+        self._builder = builder
+
+    def wid(self, name):
+        widget = self._builder.get_object(name)
+        assert(widget)
+        return widget
+
+    # push button
+    def bind_button(self, name, callback):
+        w = self.wid(name)
+        w.connect("clicked", callback)
+
+    # spin button
+    def bind_spin(self, name, config_object, key):
+        w = self.wid(name)
+        w.set_value(getattr(config_object, key))
+        w.connect("value-changed", self.bind_spin_callback, config_object, key)
+        getattr(config_object, key + '_notify_add')(w.set_value)
+
+    def bind_spin_callback(self, widget, config_object, key):
+        setattr(config_object, key, widget.get_value())
+
+    # scale
+    def bind_scale(self, name, config_object, key, widget_callback = None):
+        w = self.wid(name)
+        w.set_value(getattr(config_object, key))
+        w.connect("value-changed", self.bind_scale_callback,
+                          config_object, key, widget_callback)
+        getattr(config_object, key + '_notify_add')(w.set_value)
+
+    def bind_scale_callback(self, widget, config_object, key, callback):
+        value = widget.get_value()
+        setattr(config_object, key, value)
+        if callback:
+            callback(value)
+
+    # checkbox
+    def bind_check(self, name, config_object, key, widget_callback = None):
+        w = self.wid(name)
+        w.set_active(getattr(config_object, key))
+        if widget_callback:
+            w.connect("toggled", widget_callback, config_object, key)
+        else:
+            w.connect("toggled", self.bind_check_callback, config_object, key)
+        getattr(config_object, key + '_notify_add')(w.set_active)
+
+    def bind_check_callback(self, widget, config_object, key):
+        setattr(config_object, key, widget.get_active())
+
+    # combobox with id column
+    def bind_combobox_id(self, name, config_object, key,
+                         config_get_callback = None, config_set_callback = None):
+        w = self.wid(name)
+        if config_get_callback:
+            id = config_get_callback(config_object, key)
+        else:
+            id = str(getattr(config_object, key))
+        w.set_active_id(id)
+        w.connect("changed", self.bind_combobox_callback,
+                  config_object, key, config_set_callback)
+        notify_callback = lambda x : w.set_active_id(str(x))
+        getattr(config_object, key + '_notify_add')( \
+             lambda x: self.notify_combobox_callback(w, config_object, key,
+                                                     config_get_callback))
+    def notify_combobox_callback(self, widget,
+                                 config_object, key, config_get_callback):
+        if config_get_callback:
+            id = config_get_callback(config_object, key)
+        else:
+            id = str(getattr(config_object, key))
+        widget.set_active_id(id)
+
+    def bind_combobox_callback(self, widget,
+                               config_object, key, config_set_callback):
+        if config_set_callback:
+            config_set_callback(config_object, key, widget.get_active_id())
+        else:
+            id = widget.get_active_id()
+            assert(not id is None)  # make sure ID-Column is 0
+            if not id is None:
+                setattr(config_object, key, int(widget.get_active_id()))
+
+
+class Settings(DialogBuilder):
     def __init__(self,mainwin):
+        self.themes = {}       # cache of theme objects
 
         # Use D-bus main loop by default
-        DBusGMainLoop(set_as_default=True)
+        if "dbus" in globals():
+            dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
+        else:
+            _logger.warning("D-Bus bindings unavailable.")
 
         # finish config initialization
         config.init()
 
-        self.themes = {}       # cache of theme objects
-
+        # init dialog builder
         builder = LoadUI("settings")
-        self.window = builder.get_object("settings_window")
+        DialogBuilder.__init__(self, builder)
 
+        self.window = builder.get_object("settings_window")
         Gtk.Window.set_default_icon_name("onboard")
         self.window.set_title(_("Onboard Preferences"))
 
@@ -118,7 +216,7 @@ class Settings:
                                     self.window_state_sticky_toggle.set_active)
 
         self.force_to_top_toggle = builder.get_object("force_to_top_toggle")
-        self.force_to_top_toggle.set_active(config.window.force_to_top)
+        self.force_to_top_toggle.set_active(config.is_force_to_top())
         config.window.force_to_top_notify_add(lambda x: \
                                        [self.force_to_top_toggle.set_active(x),
                                         self.update_window_widgets()])
@@ -165,6 +263,54 @@ class Settings:
         self.inactive_transparency_delay_spinbutton.set_value(config.window.inactive_transparency_delay)
         config.window.inactive_transparency_delay_notify_add(self.inactive_transparency_delay_spinbutton.set_value)
 
+        # Keyboard - first page
+        self.bind_check("touch_feedback_enabled_toggle",
+                        config.keyboard, "touch_feedback_enabled")
+        self.bind_check("audio_feedback_enabled_toggle",
+                        config.keyboard, "audio_feedback_enabled")
+        self.bind_check("show_secondary_labels_toggle",
+                        config.keyboard, "show_secondary_labels")
+
+        # Keyboard - Advanced page
+        self.bind_combobox_id("default_key_action_combobox",
+                        config.keyboard, "default_key_action")
+        self.bind_combobox_id("key_synth_combobox",
+                        config.keyboard, "key_synth")
+
+        def get_sticky_key_behavior(config_object, key):
+            behaviors = getattr(config_object, key)
+            return behaviors.get("all", "")
+        def set_sticky_key_behavior(config_object, key, value):
+            behaviors = getattr(config_object, key).copy()
+            behaviors["all"] = value
+            setattr(config_object, key, behaviors)
+        self.bind_combobox_id("sticky_key_behavior_combobox",
+                        config.keyboard, "sticky_key_behavior",
+                        get_sticky_key_behavior, set_sticky_key_behavior)
+        self.bind_spin("sticky_key_release_delay_spinbutton",
+                            config.keyboard, "sticky_key_release_delay")
+        self.bind_combobox_id("touch_input_combobox",
+                        config.keyboard, "touch_input")
+        self.bind_combobox_id("input_event_source_combobox",
+                        config.keyboard, "input_event_source")
+
+        # word suggestions
+        self._page_word_suggestions = PageWordSuggestions(builder)
+
+        # window, docking
+        self.docking_enabled_toggle = \
+                             builder.get_object("docking_enabled_toggle")
+        self.docking_box = builder.get_object("docking_box")
+
+        def on_docking_enabled_toggle(widget, config_object, key):
+            setattr(config_object, key, widget.get_active())
+            self.update_window_widgets()
+        self.bind_check("docking_enabled_toggle",
+                        config.window, "docking_enabled",
+                        widget_callback = on_docking_enabled_toggle)
+        self.bind_button("docking_settings_button",
+                         lambda widget: DockingDialog().run(self.window))
+
         self.update_window_widgets()
 
         # layout view
@@ -172,9 +318,7 @@ class Settings:
         self.layout_view.append_column( \
                 Gtk.TreeViewColumn(None, Gtk.CellRendererText(), markup=0))
 
-        self.user_layout_root = os.path.join(config.user_dir, "layouts/")
-        if not os.path.exists(self.user_layout_root):
-            os.makedirs(self.user_layout_root)
+        self.user_layout_root = config.get_user_layout_dir()
 
         self.layout_remove_button = \
                              builder.get_object("layout_remove_button")
@@ -192,10 +336,6 @@ class Settings:
         self.customize_theme_button = \
                                    builder.get_object("customize_theme_button")
 
-        user_theme_root = Theme.user_path()
-        if not os.path.exists(user_theme_root):
-            os.makedirs(user_theme_root)
-
         self.update_themeList()
         config.theme_notify_add(self.on_theme_changed)
 
@@ -203,9 +343,8 @@ class Settings:
                     builder.get_object("system_theme_tracking_enabled_toggle")
         self.system_theme_tracking_enabled_toggle.set_active( \
                                         config.system_theme_tracking_enabled)
-        config.system_theme_tracking_enabled_notify_add(lambda x: \
-                    [self.system_theme_tracking_enabled_toggle.set_active(x),
-                     config.update_theme_from_system_theme()])
+        config.system_theme_tracking_enabled_notify_add( \
+                                        self.on_system_theme_tracking_changed)
 
         # Snippets
         self.snippet_view = SnippetView()
@@ -230,23 +369,33 @@ class Settings:
         config.universal_access.enable_click_type_window_on_exit_notify_add( \
                       self.enable_click_type_window_on_exit_toggle.set_active)
 
-        self.enable_click_type_window_on_exit_toggle = \
-                builder.get_object("enable_click_type_window_on_exit_toggle")
-        self.enable_click_type_window_on_exit_toggle.set_active( \
-                      config.universal_access.enable_click_type_window_on_exit)
-        config.universal_access.enable_click_type_window_on_exit_notify_add( \
-                      self.enable_click_type_window_on_exit_toggle.set_active)
-
         self.num_resize_handles_combobox = \
                          builder.get_object("num_resize_handles_combobox")
         self.update_num_resize_handles_combobox()
         config.resize_handles_notify_add( \
                             lambda x: self.select_num_resize_handles())
 
+        if config.mousetweaks:
+            self.bind_spin("hover_click_delay_spinbutton",
+                            config.mousetweaks, "dwell_time")
+            self.bind_spin("hover_click_motion_threshold_spinbutton",
+                            config.mousetweaks, "dwell_threshold")
+
+        # select last active page
+        page = config.current_settings_page
         self.settings_notebook = builder.get_object("settings_notebook")
-        self.settings_notebook.set_current_page(config.current_settings_page)
+        self.settings_notebook.set_current_page(page)
+
+        self.pages_view = builder.get_object("pages_view")
+        sel = self.pages_view.get_selection()
+        if sel:
+            sel.select_path(Gtk.TreePath(page))
+
         self.window.show_all()
-        #self.modeless_gksu_toggle.hide() # hidden until gksu moves to gsettings
+
+        # disable hover click controls if mousetweaks isn't installed
+        frame = builder.get_object("hover_click_frame")
+        frame.set_sensitive(bool(config.mousetweaks))
 
         self.window.set_keep_above(not mainwin)
 
@@ -256,6 +405,15 @@ class Settings:
         _logger.info("Entering mainloop of Onboard-settings")
         Gtk.main()
 
+    def on_pages_view_cursor_changed(self, widget):
+        sel = widget.get_selection()
+        if sel:
+            paths = sel.get_selected_rows()[1]
+            if paths:
+                page_num = paths[0].get_indices()[0]
+                config.current_settings_page = page_num
+                self.settings_notebook.set_current_page(page_num)
+
     def on_settings_notebook_switch_page(self, widget, gpage, page_num):
         config.current_settings_page = page_num
 
@@ -267,6 +425,14 @@ class Settings:
         _logger.info("Snippet remove button clicked")
         self.snippet_view.remove_selected()
 
+    def on_auto_show_toggled(self, widget):
+        active = widget.get_active()
+        if active and \
+           not config.check_gnome_accessibility(self.window):
+            active = False
+        config.auto_show.enabled = active
+        self.update_window_widgets()
+
     def on_status_icon_toggled(self,widget):
         config.show_status_icon = widget.get_active()
         self.update_window_widgets()
@@ -289,41 +455,46 @@ class Settings:
         config.show_tooltips = widget.get_active()
 
     def on_window_decoration_toggled(self, widget):
-        if not config.window.force_to_top:
+        if not config.is_force_to_top():
             config.window.window_decoration = widget.get_active()
         self.update_window_widgets()
 
     def on_window_state_sticky_toggled(self, widget):
-        if not config.window.force_to_top:
+        if not config.is_force_to_top():
             config.window.window_state_sticky = widget.get_active()
 
-    def on_auto_show_toggled(self, widget):
-        active = widget.get_active()
-        if active and \
-           not config.check_gnome_accessibility(self.window):
-            active = False
-        config.auto_show.enabled = active
+    def on_force_to_top_toggled(self, widget):
+        if not config.is_docking_enabled():
+            config.window.force_to_top = widget.get_active()
         self.update_window_widgets()
 
+    def on_keep_aspect_ratio_toggled(self,widget):
+        config.window.keep_aspect_ratio = widget.get_active()
+
     def update_window_widgets(self):
+        force_to_top =  config.is_force_to_top()
+
         self.icon_palette_toggle.set_sensitive( \
                              not config.is_icon_palette_last_unhide_option())
         active = config.is_icon_palette_in_use()
         if self.icon_palette_toggle.get_active() != active:
             self.icon_palette_toggle.set_active(active)
 
-        self.window_decoration_toggle.set_sensitive( \
-                                        not config.window.force_to_top)
+        self.window_decoration_toggle.set_sensitive(not force_to_top)
         active = config.has_window_decoration()
         if self.window_decoration_toggle.get_active() != active:
             self.window_decoration_toggle.set_active(active)
 
-        self.window_state_sticky_toggle.set_sensitive( \
-                                        not config.window.force_to_top)
+        self.window_state_sticky_toggle.set_sensitive( not force_to_top)
         active = config.get_sticky_state()
         if self.window_state_sticky_toggle.get_active() != active:
             self.window_state_sticky_toggle.set_active(active)
 
+        self.force_to_top_toggle.set_sensitive(not config.is_docking_enabled())
+        active = force_to_top
+        if self.force_to_top_toggle.get_active() != active:
+            self.force_to_top_toggle.set_active(active)
+
         self.background_transparency_spinbutton.set_sensitive( \
                                         not config.has_window_decoration())
         self.start_minimized_toggle.set_sensitive(\
@@ -336,12 +507,8 @@ class Settings:
         if self.enable_inactive_transparency_toggle.get_active() != active:
             self.enable_inactive_transparency_toggle.set_active(active)
 
-    def on_force_to_top_toggled(self, widget):
-        config.window.force_to_top = widget.get_active()
-        self.update_window_widgets()
-
-    def on_keep_aspect_ratio_toggled(self,widget):
-        config.window.keep_aspect_ratio = widget.get_active()
+    def update_all_widgets(self):
+        pass
 
     def on_transparent_background_toggled(self, widget):
         config.window.transparent_background = widget.get_active()
@@ -477,7 +644,7 @@ class Settings:
         if response == Gtk.ResponseType.OK:
             filename = chooser.get_filename()
 
-            f = open(filename, encoding="UTF-8")
+            f = open_utf8(filename)
             sokdoc = minidom.parse(f).documentElement
             for p in sokdoc.getElementsByTagName("pane"):
                 fn = p.attributes['filename'].value
@@ -508,7 +675,7 @@ class Settings:
 
         layouts = []
         for filename in filenames:
-            file_object = open(filename, encoding="UTF-8")
+            file_object = open_utf8(filename)
             try:
                 sokdoc = minidom.parse(file_object).documentElement
 
@@ -542,8 +709,11 @@ class Settings:
                                          os.access(filename, os.W_OK))
 
     def find_layouts(self, path):
-        files = os.listdir(path)
         layouts = []
+        try:
+            files = os.listdir(path)
+        except OSError:
+            files = []
         for filename in files:
             if filename.endswith(".sok") or \
                filename.endswith(config.LAYOUT_FILE_EXTENSION):
@@ -681,9 +851,15 @@ class Settings:
 
         self.update_themeList()
 
-    def on_theme_changed(self, theme_filename):
+    def on_system_theme_tracking_changed(self, x):
+        self.system_theme_tracking_enabled_toggle.set_active( \
+                                        config.system_theme_tracking_enabled)
+        config.load_theme()
+        self.on_theme_changed()
+
+    def on_theme_changed(self, theme_filename = None):
         selected = self.get_selected_theme_filename()
-        if selected != theme_filename:
+        if selected != config.theme_filename:
             self.update_themeList()
 
     def update_themeList(self):
@@ -745,7 +921,80 @@ class Settings:
         return None
 
 
-class ThemeDialog:
+class PageWordSuggestions(DialogBuilder):
+    """ Word Suggestions """
+
+    def __init__(self, builder):
+        DialogBuilder.__init__(self, builder)
+
+        self.wid("enable_word_suggestions_toggle") \
+                .connect_after("toggled", lambda x: self._update_ui())
+        self.bind_check("enable_word_suggestions_toggle",
+                        config.word_suggestions, "enabled")
+        self.bind_check("auto_learn_toggle",
+                            config.wp, "auto_learn")
+        self.bind_check("punctuation_assistence_toggle",
+                            config.wp, "punctuation_assistance")
+        self.bind_check("auto_capitalization_toggle",
+                            config.typing_assistance, "auto_capitalization")
+        self.bind_check("auto_correction_toggle",
+                            config.typing_assistance, "auto_correction")
+        self.bind_check("enable_spell_check_toggle",
+                            config.word_suggestions, "spelling_suggestions_enabled")
+        #self.bind_check("show_context_line_toggle",
+        #                config.word_suggestions, "show_context_line")
+        self._init_spell_checker_backend_combo()
+
+        self._update_ui()
+
+    def _init_spell_checker_backend_combo(self):
+        combo = self.wid("spell_check_backend_combobox")
+        combo.set_active(config.typing_assistance.spell_check_backend)
+        combo.connect("changed", self.on_spell_check_backend_changed)
+        config.typing_assistance.spell_check_backend_notify_add(self._backend_notify)
+
+    def on_spell_check_backend_changed(self, widget):
+        config.typing_assistance.spell_check_backend = widget.get_active()
+
+    def _backend_notify(self, mode):
+        self.wid("spell_check_backend_combobox").set_active(mode)
+
+    def _update_ui(self):
+        self.wid("word_suggestions_general_box1") \
+                .set_sensitive(config.are_word_suggestions_enabled())
+
+
+class DockingDialog(DialogBuilder):
+    """ Dialog "Docking Settings" """
+
+    def __init__(self):
+
+        builder = LoadUI("settings_docking_dialog")
+
+        DialogBuilder.__init__(self, builder)
+
+        self.bind_check("docking_shrink_workarea_toggle",
+                        config.window, "docking_shrink_workarea")
+        self.bind_check("landscape_dock_expand_toggle",
+                        config.window.landscape, "dock_expand")
+        self.bind_check("portrait_dock_expand_toggle",
+                        config.window.portrait, "dock_expand")
+        self.bind_combobox_id("docking_edge_combobox",
+                        config.window, "docking_edge")
+
+        self.update_ui()
+
+    def run(self, parent):
+        dialog = self.wid("dialog")
+        dialog.set_transient_for(parent)
+        dialog.run()
+        dialog.destroy()
+
+    def update_ui(self):
+        pass
+
+
+class ThemeDialog(DialogBuilder):
     """ Customize theme dialog """
 
     current_page = 0
@@ -756,6 +1005,7 @@ class ThemeDialog:
         self.theme = copy.deepcopy(theme)
 
         builder = LoadUI("settings_theme_dialog")
+        DialogBuilder.__init__(self, builder)
 
         self.dialog = builder.get_object("customize_theme_dialog")
 
@@ -789,6 +1039,13 @@ class ThemeDialog:
                                             "superkey_label_size_checkbutton")
         self.superkey_label_model = builder.get_object("superkey_label_model")
 
+        def on_key_stroke_width_value_changed(value):
+            self.theme.key_stroke_width = value
+            self.update_sensivity()
+        self.bind_scale("key_stroke_width_scale",
+                        config.theme_settings, "key_stroke_width",
+                        widget_callback = on_key_stroke_width_value_changed)
+
         self.update_ui()
 
         self.dialog.set_transient_for(settings.window)
@@ -1086,19 +1343,20 @@ class ThemeDialog:
         checked = self.superkey_label_size_checkbutton.get_active()
         size_group = config.SUPERKEY_SIZE_GROUP if checked else ""
         self.theme.set_superkey_label(label, size_group)
-        config.theme_settings.key_label_overrides = self.theme.key_label_overrides
+        config.theme_settings.key_label_overrides = \
+                                          dict(self.theme.key_label_overrides)
 
     def on_theme_notebook_switch_page(self, widget, gpage, page_num):
         ThemeDialog.current_page = page_num
 
 
-class ScannerDialog(object):
+class ScannerDialog(DialogBuilder):
     """ Scanner settings dialog """
 
     """ Input device columns """
     COL_ICON_NAME   = 0
     COL_DEVICE_NAME = 1
-    COL_DEVICE_INFO = 2
+    COL_DEVICE      = 2
 
     """ Device mapping columns """
     COL_NAME    = 0
@@ -1129,10 +1387,11 @@ class ScannerDialog(object):
 
     def __init__(self):
 
-        self.builder = LoadUI("settings_scanner_dialog")
-        self.wid = self.builder.get_object
+        builder = LoadUI("settings_scanner_dialog")
+        DialogBuilder.__init__(self, builder)
 
-        self.devices = Devices(self._on_device_event)
+        self.device_manager = XIDeviceManager()
+        self.device_manager.connect("device-event", self._on_device_event)
         self.pointer_selected = None
         self.mapping_renderer = None
 
@@ -1141,16 +1400,17 @@ class ScannerDialog(object):
         self.init_scan_modes()
         self.init_device_mapping()
 
-        self.bind_spin("cycles", "cycles")
-        self.bind_spin("cycles_overscan", "cycles")
-        self.bind_spin("cycles_stepscan", "cycles")
-        self.bind_spin("step_interval", "interval")
-        self.bind_spin("backtrack_interval", "interval")
-        self.bind_spin("forward_interval", "interval_fast")
-        self.bind_spin("backtrack_steps", "backtrack")
-        self.bind_check("user_scan", "user_scan")
-        self.bind_check("alternate", "alternate")
-        self.bind_check("device_detach", "device_detach")
+        scanner = config.scanner
+        self.bind_spin("cycles", scanner, "cycles")
+        self.bind_spin("cycles_overscan", scanner, "cycles")
+        self.bind_spin("cycles_stepscan", scanner, "cycles")
+        self.bind_spin("step_interval", scanner, "interval")
+        self.bind_spin("backtrack_interval", scanner, "interval")
+        self.bind_spin("forward_interval", scanner, "interval_fast")
+        self.bind_spin("backtrack_steps", scanner, "backtrack")
+        self.bind_check("user_scan", scanner, "user_scan")
+        self.bind_check("alternate", scanner, "alternate")
+        self.bind_check("device_detach", scanner, "device_detach")
 
     def __del__(self):
         _logger.debug("ScannerDialog.__del__()")
@@ -1162,7 +1422,7 @@ class ScannerDialog(object):
         dialog.destroy()
 
         config.scanner.disconnect_notifications()
-        self.devices = None
+        self.device_manager = None
 
     def init_scan_modes(self):
         combo = self.wid("scan_mode_combo")
@@ -1199,11 +1459,13 @@ class ScannerDialog(object):
 
         model.append(["input-mouse", ScanDevice.DEFAULT_NAME, None])
 
-        for dev in filter(lambda x: ScanDevice.is_pointer(x), devices):
-            model.append(["input-mouse", dev[ScanDevice.NAME], dev])
+        for dev in devices:
+            if dev.is_pointer():
+                model.append(["input-mouse", dev.name, dev])
 
-        for dev in filter(lambda x: not ScanDevice.is_pointer(x), devices):
-            model.append(["input-keyboard", dev[ScanDevice.NAME], dev])
+        for dev in devices:
+            if not dev.is_pointer():
+                model.append(["input-keyboard", dev.name, dev])
 
         self.select_current_device(config.scanner.device_name)
 
@@ -1220,9 +1482,9 @@ class ScannerDialog(object):
             combo.set_active_iter(it)
         else:
             while it:
-                info = model.get_value(it, self.COL_DEVICE_INFO)
-                if info and name == ScanDevice.get_config_string(info):
-                    self.pointer_selected = ScanDevice.is_pointer(info)
+                device = model.get_value(it, self.COL_DEVICE)
+                if device and name == device.get_config_string():
+                    self.pointer_selected = device.is_pointer()
                     self.wid("device_detach").set_sensitive(True)
                     combo.set_active_iter(it)
                     break
@@ -1239,12 +1501,12 @@ class ScannerDialog(object):
             return
 
         config.scanner.device_detach = False
-        info = model.get_value(it, self.COL_DEVICE_INFO)
+        device = model.get_value(it, self.COL_DEVICE)
 
-        if info:
-            config.scanner.device_name = ScanDevice.get_config_string(info)
+        if device:
+            config.scanner.device_name = device.get_config_string()
             self.wid("device_detach").set_sensitive(True)
-            self.pointer_selected = ScanDevice.is_pointer(info)
+            self.pointer_selected = device.is_pointer()
         else:
             config.scanner.device_name = ScanDevice.DEFAULT_NAME
             self.wid("device_detach").set_sensitive(False)
@@ -1311,10 +1573,10 @@ class ScannerDialog(object):
 
         if pointer_mode:
             col = self.COL_BUTTON
-            dev_map = config.scanner.device_button_map
+            dev_map = config.scanner.device_button_map.copy()
         else:
             col = self.COL_KEY
-            dev_map = config.scanner.device_key_map
+            dev_map = config.scanner.device_key_map.copy()
 
         dup_it = model.get_iter_from_string("0:0")
         dup_val = None
@@ -1327,7 +1589,7 @@ class ScannerDialog(object):
 
         model.set(it, col, value)
 
-        if dup_val and dev_map.has_key(dup_val):
+        if dup_val in dev_map:
             del dev_map[dup_val]
 
         action = model.get_value(it, self.COL_ACTION)
@@ -1350,25 +1612,27 @@ class ScannerDialog(object):
             return
 
         if pointer_mode:
-            old_value = model.get(it, self.COL_BUTTON)
+            old_value = model.get_value(it, self.COL_BUTTON)
             model.set(it, self.COL_BUTTON, 0)
-            dev_map = config.scanner.device_button_map
-            if dev_map.has_key(old_value):
-                del dev_map[old_value]
-                config.scanner.device_button_map = dev_map
+            if old_value in config.scanner.device_button_map:
+                copy = config.scanner.device_button_map.copy()
+                del copy[old_value]
+                config.scanner.device_button_map = copy
         else:
-            old_value = model.get(it, self.COL_KEY)
+            old_value = model.get_value(it, self.COL_KEY)
             model.set(it, self.COL_KEY, 0)
-            dev_map = config.scanner.device_key_map
-            if dev_map.has_key(old_value):
-                del dev_map[old_value]
-                config.scanner.device_key_map = dev_map
+            if old_value in config.scanner.device_key_map:
+                copy = config.scanner.device_key_map.copy()
+                del copy[old_value]
+                config.scanner.device_key_map = copy
 
     def list_devices(self):
-        return filter(ScanDevice.is_useable, self.devices.list())
+        return [d for d in self.device_manager.get_devices() \
+                if ScanDevice.is_useable(d) ]
 
-    def _on_device_event(self, event, device_id, detail):
-        if event in ["DeviceAdded", "DeviceRemoved"]:
+    def _on_device_event(self, event):
+        if event.xi_type in [XIEventType.DeviceAdded,
+                             XIEventType.DeviceRemoved]:
             self.update_input_devices()
 
     def get_value_for_action(self, action, dev_map):
@@ -1376,24 +1640,6 @@ class ScannerDialog(object):
             if v == action:
                 return k
 
-    def bind_spin(self, name, key):
-        w = self.wid(name)
-        w.set_value(getattr(config.scanner, key))
-        w.connect("value-changed", self.bind_spin_callback, key)
-        getattr(config.scanner, key + '_notify_add')(w.set_value)
-
-    def bind_spin_callback(self, widget, key):
-        setattr(config.scanner, key, widget.get_value())
-
-    def bind_check(self, name, key):
-        w = self.wid(name)
-        w.set_active(getattr(config.scanner, key))
-        w.connect("toggled", self.bind_check_callback, key)
-        getattr(config.scanner, key + '_notify_add')(w.set_active)
-
-    def bind_check_callback(self, widget, key):
-        setattr(config.scanner, key, widget.get_active())
-
 
 MAX_GINT32 = (1 << 31) - 1
 
@@ -1525,7 +1771,10 @@ class CellRendererMapping(Gtk.CellRendererText):
         return self._sizing_label.get_preferred_width()
 
     def do_start_editing(self, event, widget, path, bg_area, cell_area, state):
+        if not event: # else SEGFAULT when pressing a keyboard key twice
+            return
 
+        time = event.get_time()
         device = event.get_device()
         if device.get_source() == Gdk.InputSource.KEYBOARD:
             keyboard = device
@@ -1537,13 +1786,13 @@ class CellRendererMapping(Gtk.CellRendererText):
         if keyboard.grab(widget.get_window(),
                          Gdk.GrabOwnership.WINDOW, False,
                          Gdk.EventMask.KEY_PRESS_MASK,
-                         None, event.get_time()) != Gdk.GrabStatus.SUCCESS:
+                         None, time) != Gdk.GrabStatus.SUCCESS:
             return
 
         if pointer.grab(widget.get_window(),
                         Gdk.GrabOwnership.WINDOW, False,
                         Gdk.EventMask.BUTTON_PRESS_MASK,
-                        None, event.get_time()) != Gdk.GrabStatus.SUCCESS:
+                        None, time) != Gdk.GrabStatus.SUCCESS:
             keyboard.ungrab(time)
             return
 
