@@ -13,7 +13,7 @@ from subprocess import Popen
 from math import pi, sin, cos, exp, sqrt, log
 from contextlib import contextmanager
 
-from gi.repository import GObject, Gtk
+from gi.repository import GLib, Gtk
 
 ### Logging ###
 import logging
@@ -21,25 +21,49 @@ from functools import reduce
 _logger = logging.getLogger("utils")
 ###############
 
+
+# keycodes
+class KeyCode:
+    Return   = 36
+    KP_Enter = 104
+    C        = 54
+
+class Modifiers:
+    # 1      2     4    8    16     32    64     128
+    SHIFT, CAPS, CTRL, ALT, NUMLK, MOD3, SUPER, ALTGR = \
+               (1<<bit for bit in range(8))
+
+# modifiers affecting labels
+LABEL_MODIFIERS = Modifiers.SHIFT | \
+                  Modifiers.CAPS | \
+                  Modifiers.NUMLK | \
+                  Modifiers.ALTGR
+
 modifiers = {"shift":1,
              "caps":2,
              "control":4,
-             "mod1":8,
-             "mod2":16,
+             "mod1":8,   # Left Alt
+             "mod2":16,  # NumLk
              "mod3":32,
-             "mod4":64,
-             "mod5":128}
+             "mod4":64,  # Super
+             "mod5":128, # Alt Gr
+            }
 
+modGroups = {"SHIFT" : ["LFSH"],
+             "CTRL"  : ["LCTL"],
+            }
 
-modDic = {"LWIN" : ("Win",64),
-          "RTSH" : ("⇧", 1),
-          "LFSH" : ("⇧", 1),
-          "RALT" : ("Alt Gr", 128),
-          "LALT" : ("Alt", 8),
-          "RCTL" : ("Ctrl", 4),
-          "LCTL" : ("Ctrl", 4),
-          "CAPS" : ("CAPS", 2),
-          "NMLK" : ("Nm\nLk",16)}
+modList = [["LWIN", ("Win",64)],
+           ["RTSH", ("⇧", 1)],
+           ["LFSH", ("⇧", 1)],
+           ["RALT", ("Alt Gr", 128)],
+           ["LALT", ("Alt", 8)],
+           ["RCTL", ("Ctrl", 4)],
+           ["LCTL", ("Ctrl", 4)],
+           ["CAPS", ("CAPS", 2)],
+           ["NMLK", ("Nm\nLk",16)]]
+
+modDic = dict(modList)
 
 otherDic = {"RWIN" : "Win",
             "MENU" : "Menu",
@@ -103,6 +127,69 @@ keysyms = {"space" : 65408,
 def get_keysym_from_name(name):
     return keysyms[name]
 
+def parse_key_combination(combo, avaliable_key_ids = None):
+    """
+    Parses a key combination into a list of modifier masks and key_ids.
+    The key-id part of the combo may contain a regex pattern.
+
+    Doctests:
+
+    # modifiers
+    >>> parse_key_combination(["TAB"], ["TAB"])
+    [('TAB', 0)]
+    >>> parse_key_combination(["LALT", "TAB"], ["TAB"])
+    [('TAB', 8)]
+    >>> parse_key_combination(["LALT", "LFSH", "TAB"], ["TAB"])
+    [('TAB', 9)]
+    >>> parse_key_combination(["LWIN", "RTSH", "LFSH", "RALT", "LALT", "RCTL", "LCTL", "CAPS", "NMLK", "TAB"], ["TAB"])
+    [('TAB', 223)]
+
+    # modifier groups
+    >>> parse_key_combination(["CTRL", "SHIFT", "TAB"], ["TAB"])
+    [('TAB', 5)]
+
+    # regex
+    >>> parse_key_combination(["F\d+"], ["TAB", "F1", "F2", "F3", "F9"])
+    [('F1', 0), ('F2', 0), ('F3', 0), ('F9', 0)]
+    """
+    modifiers = combo[:-1]
+    key_pattern = combo[-1]
+
+    # find modifier mask
+    mod_mask = parse_modifier_strings(modifiers)
+    if mod_mask is None:
+        return None
+
+    # match regex key id with all available ids
+    results = []
+    pattern = re.compile(key_pattern)
+    for key_id in avaliable_key_ids:
+        match = pattern.match(key_id)
+        if match and match.group() == key_id:
+            results.append((key_id, mod_mask))
+
+    return results
+
+def parse_modifier_strings(modifiers):
+    """ Build modifier mask from modifier strings. """
+    mod_mask = 0
+    for modifier in modifiers:
+        m = modDic.get(modifier)
+        if not m is None:
+            mod_mask |= m[1]
+        else:
+            group = modGroups.get(modifier)
+            if not group is None:
+                for mod in group:
+                    mod_mask |= modDic[mod][1]
+            else:
+                _logger.warning("unrecognized modifier '{}'; try one of {}" \
+                                .format(modifier, ",".join(m[0] for m in modList)))
+                mod_mask = None
+                break
+
+    return mod_mask
+
 def run_script(script):
     a =__import__(script)
     a.run()
@@ -194,14 +281,14 @@ class dictproperty(object):
 def show_error_dialog(error_string):
     """ Show an error dialog """
 
-    error_dlg = Gtk.MessageDialog(type=Gtk.MessageType.ERROR,
+    error_dlg = Gtk.MessageDialog(message_type=Gtk.MessageType.ERROR,
                                   message_format=error_string,
                                   buttons=Gtk.ButtonsType.OK)
     error_dlg.run()
     error_dlg.destroy()
 
 def show_ask_string_dialog(question, parent=None):
-    question_dialog = Gtk.MessageDialog(type=Gtk.MessageType.QUESTION,
+    question_dialog = Gtk.MessageDialog(message_type=Gtk.MessageType.QUESTION,
                                         buttons=Gtk.ButtonsType.OK_CANCEL)
     if parent:
         question_dialog.set_transient_for(parent)
@@ -220,7 +307,7 @@ def show_confirmation_dialog(question, parent=None):
     """
     Show this dialog to ask confirmation before executing a task.
     """
-    dlg = Gtk.MessageDialog(type=Gtk.MessageType.QUESTION,
+    dlg = Gtk.MessageDialog(message_type=Gtk.MessageType.QUESTION,
                             message_format=question,
                             buttons=Gtk.ButtonsType.YES_NO)
     if parent:
@@ -233,7 +320,7 @@ def show_new_device_dialog(name, config_string, is_pointer, callback):
     """
     Show a "New Input Device" dialog.
     """
-    dialog = Gtk.MessageDialog(type  = Gtk.MessageType.OTHER,
+    dialog = Gtk.MessageDialog(message_type  = Gtk.MessageType.OTHER,
                                title = _("New Input Device"),
                                text  = _("Onboard has detected a new input device"))
     if is_pointer:
@@ -331,7 +418,7 @@ def merge_tuple_strings(text1, text2):
 
 class CallOnce(object):
     """
-    call each <callback> during <delay> only once
+    Call each <callback> during <delay> only once
     Useful to reduce a storm of config notifications
     to just a single (or a few) update(s) of onboards state.
     """
@@ -342,6 +429,9 @@ class CallOnce(object):
         self.delay = delay
         self.delay_forever = delay_forever
 
+    def is_running(self):
+        return not self.timer is None
+
     def enqueue(self, callback, *args):
         if not callback in self.callbacks:
             self.callbacks[callback] = args
@@ -350,11 +440,17 @@ class CallOnce(object):
             pass
 
         if self.delay_forever and self.timer:
-            GObject.source_remove(self.timer)
+            GLib.source_remove(self.timer)
             self.timer = None
 
         if not self.timer and self.callbacks:
-            self.timer = GObject.timeout_add(self.delay, self.cb_timer)
+            self.timer = GLib.timeout_add(self.delay, self.cb_timer)
+
+    def stop(self):
+        if self.timer:
+            GLib.source_remove(self.timer)
+            self.timer = None
+        self.callbacks.clear()
 
     def cb_timer(self):
         for callback, args in list(self.callbacks.items()):
@@ -363,8 +459,7 @@ class CallOnce(object):
             except:
                 traceback.print_exc()
 
-        self.callbacks.clear()
-        self.timer = None
+        self.stop()
         return False
 
 
@@ -486,6 +581,11 @@ class Rect:
     def int(self):
         return Rect(int(self.x), int(self.y), int(self.w), int(self.h))
 
+    def scale(self, kx, ky = None):
+        if ky == None:
+            ky = kx
+        return Rect(self.x * kx, self.y * ky, self.w * kx, self.h * ky)
+
     def offset(self, dx, dy):
         """
         Returns a new Rect displaced by dx and dy.
@@ -528,7 +628,24 @@ class Rect:
                     w, h)
 
     def intersects(self, rect):
-        return not self.intersection(rect).is_empty()
+        """
+        Doctests:
+        >>> Rect(0, 0, 1, 1).intersects(Rect(0, 0, 1, 1))
+        True
+        >>> Rect(0, 0, 1, 1).intersects(Rect(1, 0, 1, 1))
+        False
+        >>> Rect(1, 0, 1, 1).intersects(Rect(0, 0, 1, 1))
+        False
+        >>> Rect(0, 0, 1, 1).intersects(Rect(0, 1, 1, 1))
+        False
+        >>> Rect(0, 1, 1, 1).intersects(Rect(0, 0, 1, 1))
+        False
+        """
+        #return not self.intersection(rect).is_empty()
+        return not (self.x >= rect.x + rect.w or \
+                    self.x + self.w <= rect.x or \
+                    self.y >= rect.y + rect.h or \
+                    self.y + self.h <= rect.y)
 
     def intersection(self, rect):
        x0 = max(self.x, rect.x)
@@ -568,15 +685,24 @@ class Rect:
 
     def align_rect(self, rect, x_align = 0.5, y_align = 0.5):
         """
-        Alignes the given rect inside of self.
+        Aligns the given rect inside of self.
         x/y_align = 0.5 centers rect.
         """
         x = self.x + (self.w - rect.w) * x_align
         y = self.y + (self.h - rect.h) * y_align
-        return Rect(x, y, rect.w, rect. h)
+        return Rect(x, y, rect.w, rect.h)
 
-    def subdivide(self, rows, columns, x_spacing = None, y_spacing = None):
-        """ Divide self into rows x columns sub-rectangles """
+    def align_at_point(self, x, y, x_align = 0.5, y_align = 0.5):
+        """
+        Aligns the given rect to a point.
+        x/y_align = 0.5 centers rect.
+        """
+        x = x - self.w * x_align
+        y = y - self.h * y_align
+        return Rect(x, y, self.w, self.h)
+
+    def subdivide(self, columns, rows, x_spacing = None, y_spacing = None):
+        """ Divide self into columns x rows sub-rectangles """
         if y_spacing is None:
             y_spacing = x_spacing
         if x_spacing is None:
@@ -643,6 +769,45 @@ def roundrect_arc(context, rect, r = 15):
 
 
 def roundrect_curve(context, rect, r_pct = 100):
+    """
+    Uses B-splines for less even looks than with arcs, but
+    still allows for approximate circles at r_pct = 100.
+    """
+    x0 = rect.x
+    y0 = rect.y
+    w  = rect.w
+    h  = rect.h
+    x1 = x0 + w
+    y1 = y0 + h
+
+    r = min(w, h) * min(r_pct/100.0, 0.5) # full range at 50%
+    k = (r-1) * r_pct/200.0 # position of control points for circular curves
+
+    line_to = context.line_to
+    curve_to = context.curve_to
+
+    # top left
+    context.move_to(x0+r, y0)
+
+    # top right
+    line_to(x1-r,y0)
+    curve_to(x1-k, y0, x1, y0+k, x1, y0+r)
+
+    # bottom right
+    line_to(x1, y1-r)
+    curve_to(x1, y1-k, x1-k, y1, x1-r, y1)
+
+    # bottom left
+    line_to(x0+r, y1)
+    curve_to(x0+k, y1, x0, y1-k, x0, y1-r)
+
+    # top left
+    line_to(x0, y0+r)
+    curve_to(x0, y0+k, x0+k, y0, x0+r, y0)
+
+    context.close_path ()
+
+def roundrect_curve_old(context, rect, r_pct = 100):
     """
     Uses B-splines for less even looks than with arcs, but
     still allows for approximate circles at r_pct = 100.
@@ -717,7 +882,7 @@ def gradient_line(rect, alpha):
             r * sin(alpha) + y0 + b,
            -r * cos(alpha) + x0 + a,
            -r * sin(alpha) + y0 + b)
-
+import cairo
 def drop_shadow(cr, pattern, bounds, blur_radius = 4.0, offset = (0, 0),
                                   alpha=0.06, steps=4):
     """
@@ -725,6 +890,7 @@ def drop_shadow(cr, pattern, bounds, blur_radius = 4.0, offset = (0, 0),
     damage rects smaller than the full window rect.
     """
     origin = bounds.get_center()
+    cr.set_source_rgba(0.0, 0.0, 0.0, alpha)
     for i in range(steps):
 
         x = (i if i else 0.5) / float(steps)
@@ -737,7 +903,6 @@ def drop_shadow(cr, pattern, bounds, blur_radius = 4.0, offset = (0, 0),
         cr.translate(*origin)
         cr.scale(x_scale, y_scale)
         cr.translate(-origin[0] + offset[0], -origin[1] + offset[1])
-        cr.set_source_rgba(0.0, 0.0, 0.0, alpha)
 
         cr.mask(pattern)
         cr.restore()
@@ -763,7 +928,7 @@ def timeit(s, out=sys.stdout):
 
 class Timer(object):
     """
-    Simple wrapper around gobject's timer API
+    Simple wrapper around GLib's timer API
     Overload on_timer in derived classes.
     For one-shot timers return False there.
     """
@@ -791,10 +956,10 @@ class Timer(object):
         self.stop()
 
         if type(delay) == int:
-            self._timer = GObject.timeout_add_seconds(delay, self._cb_timer)
+            self._timer = GLib.timeout_add_seconds(delay, self._cb_timer)
         else:
             ms = int(delay * 1000.0)
-            self._timer = GObject.timeout_add(ms, self._cb_timer)
+            self._timer = GLib.timeout_add(ms, self._cb_timer)
 
     def finish(self):
         """
@@ -806,7 +971,7 @@ class Timer(object):
 
     def stop(self):
         if self.is_running():
-            GObject.source_remove(self._timer)
+            GLib.source_remove(self._timer)
             self._timer = None
 
     def is_running(self):
@@ -828,6 +993,16 @@ class Timer(object):
         return True
 
 
+class TimerOnce(Timer):
+    def on_timer(self):
+        """
+        Overload this.
+        """
+        if self._callback:
+            return self._callback(*self._callback_args)
+        return False
+
+
 class DelayedLauncher(Timer):
     """
     Launches a process after a certain delay.
@@ -932,6 +1107,14 @@ class TreeItem(object):
         for item in items:
             item.parent = self
 
+    def append_item(self, item):
+        if self.items:
+            self.items.append(item)
+        else:
+            self.items = [item]
+
+        item.parent = self
+
     def append_items(self, items):
         if self.items:
             self.items += items
@@ -941,13 +1124,20 @@ class TreeItem(object):
         for item in items:
             item.parent = self
 
+    def get_parent(self):
+        return self.parent
+
     def find_ids(self, ids):
         """ find all items with matching id """
-        items = []
         for item in self.iter_items():
             if item.id in ids:
-                items.append(item)
-        return items
+                yield item
+
+    def find_classes(self, item_classes):
+        """ find all items with matching id """
+        for item in self.iter_items():
+            if isinstance(item, item_classes):
+                yield item
 
     def iter_items(self):
         """
@@ -956,7 +1146,7 @@ class TreeItem(object):
         yield self
 
         for item in self.items:
-            for child in item.iter_depth_first():
+            for child in item.iter_items():
                 yield child
 
     def iter_depth_first(self):
@@ -1037,6 +1227,13 @@ class Process:
             cmdline = f.read()
         return cmdline
 
+    @staticmethod
+    def get_process_name(pid):
+        args = Process.get_cmdline(pid).split("\0")
+        if args:
+            return os.path.basename(args[0])
+        return ""
+
     @staticmethod
     def was_launched_by(process_name):
         """ Checks if this process was launched by <process_name> """
@@ -1046,6 +1243,29 @@ class Process:
             return process_name in cmdline
         return False
 
+def exists_in_path(basename):
+    """
+    Does a file with this basename exist anywhere in PATH's directories?
+    """
+    for path in os.environ["PATH"].split(os.pathsep):
+        filename = os.path.join(path, basename)
+        if os.path.isfile(filename):
+            return True
+    return False
+
+def chmodtree(path, mode = 0o777, only_dirs = False):
+    """
+    Change permissions of all files of the given directory tree.
+    Raises OSError.
+    """
+    os.chmod(path, mode)
+    for root, dirs, files in os.walk(path):
+        for d in dirs:
+            os.chmod(os.path.join(root, d), mode)                    
+        if not only_dirs:
+            for f in files:
+                os.chmod(os.path.join(root, f), mode)                    
+
 def unicode_str(obj, encoding = "utf-8"):
     """
     Safe str() function that always returns an unicode string.
@@ -1062,6 +1282,16 @@ def unicode_str(obj, encoding = "utf-8"):
 
     return str(obj).decode("utf-8")  # strings, numbers, ...
 
+def open_utf8(filename, mode = "r"):
+    """
+    Python 2 compatible replacement for builtin open().
+    Python 3 added the encoding parameter.
+    """
+    if sys.version_info.major == 2:
+        return open(filename, mode)
+    else:
+        return open(filename, mode=mode, encoding="UTF-8")
+
 
 class Translation:
     """
@@ -1096,7 +1326,6 @@ class Translation:
 
     def format(self, msgid, *args, **kwargs):
         """ Safe replacement for str.format() """
-
         msgstr = self.ugettext(msgid)
         try:
             result = msgstr.format(*args, **kwargs)
@@ -1111,3 +1340,286 @@ class Translation:
         return result
 
 
+class EventSource(object):
+    """ Simple event handling based on python callbacks """
+    _event_queue = None  # for optional async delivery
+
+    def __init__(self, event_names):
+        self._callbacks = dict((e,[]) for e in event_names)
+
+    def cleanup():
+        self.flush_events()
+
+    def connect(self, event_name, callback):
+        callbacks = self._callbacks[event_name]
+        if not callback in callbacks:
+            callbacks.append(callback)
+
+    def disconnect(self, event_name, callback):
+        callbacks = self._callbacks[event_name]
+        if callback in callbacks:
+            callbacks.remove(callback)
+
+    def has_listeners(self, event_names = None):
+        """ 
+        Are there callbacks registered for the given event_names or any event?
+        """
+        if event_names:
+            return any(bool(self._callbacks[name]) for name in event_names)
+        return any(bool(value) for value in self._callbacks.values())
+
+    def emit(self, event_name, *args, **kwargs):
+        """
+        Send event, call all listener's callbacks.
+        """
+        #print("emit", event_name, list(args), kwargs)
+        for callback in self._callbacks[event_name]:
+            callback(*args, **kwargs)
+
+    def emit_async(self, event_name, *args, **kwargs):
+        """
+        Queue up asynchronous event.
+        """
+        #print("emit_async", event_name, list(args), kwargs)
+        event = (event_name, args, kwargs)
+        if self._event_queue is None:
+            self._event_queue = [event]
+            GLib.idle_add(self.flush_events)
+        else:
+            self._event_queue.append(event)
+
+    def flush_events(self):
+        """
+        Send pending asynchronous events.
+        """
+        if not self._event_queue is None:
+            for event_name, args, kwargs in self._event_queue:
+                self.emit(event_name, *args, **kwargs)
+            self.clear_events()
+
+    def clear_events(self):
+        """
+        Cancel pending asynchronous events.
+        """
+        self._event_queue = None
+
+
+def permute_mask(mask):
+    """
+    Return all permutations of the bits in mask.
+
+    Doctests:
+    >>> permute_mask(1)
+    [0, 1]
+    >>> permute_mask(5)
+    [0, 1, 4, 5]
+    >>> permute_mask(14)
+    [0, 2, 4, 6, 8, 10, 12, 14]
+    """
+    bit_masks = [bit_mask for bit_mask in (1<<bit for bit in range(8)) \
+                 if mask & bit_mask]
+    n = len(bit_masks)
+    perms = []
+    for i in range(2**n):
+        m = 0
+        for bit in range(n):
+            if i & 1<<bit:
+                m |= bit_masks[bit]
+        perms.append(m)
+    return perms
+
+
+class XDGDirs:
+    """
+    Build paths compliant with XDG Base Directory Specification.
+    http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html
+
+    Doctests:
+
+    >>> os.environ["HOME"] = "/home/test_user"
+
+    # XDG_CONFIG_HOME unavailable
+    >>> os.environ["XDG_CONFIG_HOME"] = ""
+    >>> XDGDirs.get_config_home("onboard/test.dat")
+    '/home/test_user/.config/onboard/test.dat'
+
+    # XDG_CONFIG_HOME available
+    >>> os.environ["XDG_CONFIG_HOME"] = "/home/test_user/.config_home"
+    >>> XDGDirs.get_config_home("onboard/test.dat")
+    '/home/test_user/.config_home/onboard/test.dat'
+
+    # XDG_DATA_HOME unavailable
+    >>> os.environ["XDG_DATA_HOME"] = ""
+    >>> XDGDirs.get_data_home("onboard/test.dat")
+    '/home/test_user/.local/share/onboard/test.dat'
+
+    # XDG_DATA_HOME available
+    >>> os.environ["XDG_DATA_HOME"] = "/home/test_user/.data_home"
+    >>> XDGDirs.get_data_home("onboard/test.dat")
+    '/home/test_user/.data_home/onboard/test.dat'
+
+    # XDG_CONFIG_DIRS unvailable
+    >>> os.environ["XDG_CONFIG_HOME"] = ""
+    >>> os.environ["XDG_CONFIG_DIRS"] = ""
+    >>> XDGDirs.get_all_config_dirs("onboard/test.dat")
+    ['/home/test_user/.config/onboard/test.dat', '/etc/xdg/onboard/test.dat']
+
+    # XDG_CONFIG_DIRS available
+    >>> os.environ["XDG_CONFIG_HOME"] = ""
+    >>> os.environ["XDG_CONFIG_DIRS"] = "/etc/xdg/xdg-ubuntu:/etc/xdg"
+    >>> XDGDirs.get_all_config_dirs("onboard/test.dat")
+    ['/home/test_user/.config/onboard/test.dat', \
+'/etc/xdg/xdg-ubuntu/onboard/test.dat', \
+'/etc/xdg/onboard/test.dat']
+
+    # XDG_DATA_DIRS unvailable
+    >>> os.environ["XDG_DATA_HOME"] = ""
+    >>> os.environ["XDG_DATA_DIRS"] = ""
+    >>> XDGDirs.get_all_data_dirs("onboard/test.dat")
+    ['/home/test_user/.local/share/onboard/test.dat', \
+'/usr/local/share/onboard/test.dat', \
+'/usr/share/onboard/test.dat']
+
+    # XDG_DATA_DIRS available
+    >>> os.environ["XDG_DATA_HOME"] = ""
+    >>> os.environ["XDG_DATA_DIRS"] = "/usr/share/gnome:/usr/local/share/:/usr/share/"
+    >>> XDGDirs.get_all_data_dirs("onboard/test.dat")
+    ['/home/test_user/.local/share/onboard/test.dat', \
+'/usr/share/gnome/onboard/test.dat', \
+'/usr/local/share/onboard/test.dat', \
+'/usr/share/onboard/test.dat']
+    """
+
+    @staticmethod
+    def get_config_home(file = None):
+        """
+        User specific config directory.
+        """
+        path = os.environ.get("XDG_CONFIG_HOME")
+        if path and not os.path.isabs(path):
+            _logger.warning("XDG_CONFIG_HOME doesn't contain an absolute path,"
+                            "ignoring.")
+            path = None
+        if not path:
+            path = os.path.join(os.path.expanduser("~"), ".config")
+
+        if file:
+            path = os.path.join(path, file)
+
+        return path
+
+    @staticmethod
+    def get_config_dirs():
+        """
+        Config directories ordered by preference.
+        """
+        paths = []
+
+        value = os.environ.get("XDG_CONFIG_DIRS")
+        if not value:
+            value = "/etc/xdg"
+        
+        paths = value.split(":")
+        paths = [p for p in paths if os.path.isabs(p)]
+
+        return paths
+
+    @staticmethod
+    def get_all_config_dirs(file = None):
+        paths = [XDGDirs.get_config_home()] + XDGDirs.get_config_dirs()
+
+        if file:
+            paths = [os.path.join(p, file) for p in paths]
+
+        return paths
+
+    @staticmethod
+    def find_config_files(file):
+        """ Find file in all config directories, highest priority first. """
+        paths = XDGDirs.get_all_config_dirs(file)
+        return [p for p in paths if os.path.isfile(path) and \
+                                    os.access(filename, os.R_OK)]
+
+    @staticmethod
+    def find_config_file(file):
+        """ Find file of highest priority """
+        paths = XDGDirs.find_config_files(file)
+        if paths:
+            return paths[0]
+        return None
+
+    @staticmethod
+    def get_data_home(file = None):
+        """
+        User specific data directory.
+        """
+        path = os.environ.get("XDG_DATA_HOME")
+        if path and not os.path.isabs(path):
+            _logger.warning("XDG_DATA_HOME doesn't contain an absolute path,"
+                            "ignoring.")
+            path = None
+        if not path:
+            path = os.path.join(os.path.expanduser("~"), ".local", "share")
+
+        if file:
+            path = os.path.join(path, file)
+
+        return path
+
+    @staticmethod
+    def get_data_dirs():
+        """
+        Data directories ordered by preference.
+        """
+        paths = []
+
+        value = os.environ.get("XDG_DATA_DIRS")
+        if not value:
+            value = "/usr/local/share/:/usr/share/"
+        
+        paths = value.split(":")
+        paths = [p for p in paths if os.path.isabs(p)]
+
+        return paths
+
+    @staticmethod
+    def get_all_data_dirs(file = None):
+        paths = [XDGDirs.get_data_home()] + XDGDirs.get_data_dirs()
+
+        if file:
+            paths = [os.path.join(p, file) for p in paths]
+
+        return paths
+
+    @staticmethod
+    def find_data_files(file):
+        """ Find file in all data directories, highest priority first. """
+        paths = XDGDirs.get_all_data_dirs(file)
+        return [p for p in paths if os.path.isfile(path) and \
+                                    os.access(filename, os.R_OK)]
+
+    @staticmethod
+    def find_data_file(file):
+        """ Find file of highest priority """
+        paths = XDGDirs.find_data_files(file)
+        if paths:
+            return paths[0]
+        return None
+
+    def assure_user_dir_exists(path):
+        """
+        If necessary create user XDG directory.
+        Raises OSError.
+        """
+        exists = os.path.exists(path)
+        if not exists:
+            try:
+                os.makedirs(path, mode = 0o700)
+                exists = True
+            except OSError as ex:
+                _logger.error(_format("failed to create directory '{}': {}",
+                                       path, unicode_str(ex)))
+                raise ex
+        return exists
+
+
